"""
Tests that do not need to connect servers
"""
# pylint:disable=unused-variable

from typing import Type
from django.apps.registry import Apps
from django.test import TestCase
from django.db.models import DO_NOTHING, Subquery
from salesforce import fields, models
from salesforce.dbapi import driver
from salesforce.testrunner.example.models import (
        Contact, Opportunity, OpportunityContactRole, ChargentOrder, Test as TestModel)
from salesforce.backend.test_helpers import default_is_sf, LazyTestMixin, skipUnless
from salesforce.backend.utils import sobj_id


class EasyCharField(models.CharField):
    def __init__(self, max_length=255, null=True, default='', **kwargs):
        super().__init__(max_length=max_length, null=null, default=default, **kwargs)


class EasyForeignKey(models.ForeignKey):
    def __init__(self, othermodel, on_delete=DO_NOTHING, **kwargs):
        super().__init__(othermodel, on_delete=on_delete, **kwargs)


class TestField(TestCase):
    """
    Unit tests for salesforce.fields that don't need to connect Salesforce.
    """
    def test_primary_key(self):
        """
        Verify the expected attributes of primary key
        """
        test_apps = Apps(['salesforce.testrunner.example'])

        class Ab(models.SalesforceModel):
            class Meta:
                app_label = 'example'
                apps = test_apps
        self.assertTrue(isinstance(fields.SF_PK, str))
        self.assertTrue(hasattr(Ab(), 'pk'))
        self.assertTrue(hasattr(Ab(), fields.SF_PK))

    def test_auto_db_column(self):
        """
        Verify that db_column is not important in most cases, but it has
        precedence if it is specified.
        Verify it for lower_case and CamelCase conventions, for standard fields
        and for custom fields, for normal fields and for foreign keys.
        """
        test_apps = Apps(['salesforce.testrunner.example'])

        class Aa(models.SalesforceModel):
            class Meta:
                app_label = 'example'
                apps = test_apps

        class Dest(models.SalesforceModel):
            class Meta:
                app_label = 'example'
                apps = test_apps

        def test(field, expect_attname, expect_column):
            "Compare field attributes with expected `attname` and `column`."
            field.contribute_to_class(Dest, field.name)
            self.assertEqual(field.get_attname_column(), (expect_attname, expect_column))
        # Normal fields
        test(EasyCharField(name='LastName'), 'LastName', 'LastName')
        test(EasyCharField(name='last_name'), 'last_name', 'LastName')
        test(EasyCharField(name='MyCustomField', custom=True), 'MyCustomField', 'MyCustomField__c')
        test(EasyCharField(name='MyCustomField', custom=True, db_column='UglyName__c'), 'MyCustomField', 'UglyName__c')
        test(EasyCharField(name='my_custom_field', custom=True), 'my_custom_field', 'MyCustomField__c')
        test(EasyCharField(name='Payment_Method', custom=True), 'Payment_Method', 'Payment_Method__c')
        # Foreign keys to a class Aa
        test(EasyForeignKey(Aa, name='Account'), 'AccountId', 'AccountId')
        test(EasyForeignKey(Aa, name='account'), 'account_id', 'AccountId')
        test(EasyForeignKey(Aa, name='account', db_column='AccountId'), 'account_id', 'AccountId')
        test(EasyForeignKey(Aa, name='account', db_column='UglyNameId'), 'account_id', 'UglyNameId')
        test(EasyForeignKey(Aa, name='CampaignMember'), 'CampaignMemberId', 'CampaignMemberId')
        test(EasyForeignKey(Aa, name='campaign_member'), 'campaign_member_id', 'CampaignMemberId')
        test(EasyForeignKey(Aa, name='MyCustomForeignField', custom=True),
             'MyCustomForeignFieldId',
             'MyCustomForeignField__c')
        test(EasyForeignKey(Aa, name='my_custom_foreign_field', custom=True),
             'my_custom_foreign_field_id',
             'MyCustomForeignField__c')


class TestQueryCompiler(TestCase, LazyTestMixin):
    def test_namespaces_auto(self):
        """Verify that the database column name can be correctly autodetected

        from model Meta for managed packages with a namespace prefix.
        (The package need not be installed for this unit test.)
        """
        tested_field = ChargentOrder._meta.get_field('Balance_Due')
        self.assertEqual(tested_field.sf_custom, True)
        self.assertEqual(tested_field.column, 'ChargentOrders__Balance_Due__c')

    def test_subquery_filter_on_parent(self):
        """Regression test with a filter based on subquery.

        This test is very similar to the required example in PR #103.
        """
        qs = OpportunityContactRole.objects.filter(
            role='abc',
            opportunity__in=Opportunity.objects.filter(stage='Prospecting')
        )
        sql, params = qs.query.get_compiler('salesforce').as_sql()
        self.assertRegex(sql,
                         "WHERE Opportunity.StageName =",
                         "Probably because aliases are invalid for SFDC, e.g. 'U0.StageName'")
        self.assertRegex(sql,
                         r'SELECT .*OpportunityContactRole\.Role.* '
                         r'FROM OpportunityContactRole WHERE \(.* AND .*\)')
        self.assertRegex(sql,
                         r'OpportunityContactRole.OpportunityId IN '
                         r'\(SELECT Opportunity\.Id FROM Opportunity WHERE Opportunity\.StageName = %s ?\)')
        self.assertRegex(sql, 'OpportunityContactRole.Role = %s')

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_subquery_filter_on_child(self):
        """Filter with a Subquery() on a child object.

        Especially useful with ManyToMany field relationships.
        """
        associations = OpportunityContactRole.objects.filter(contact__email='a@example.com')
        soql = ("SELECT Opportunity.Name FROM Opportunity WHERE Opportunity.Id IN ("
                "SELECT OpportunityContactRole.OpportunityId FROM OpportunityContactRole "
                "WHERE OpportunityContactRole.Contact.Email = 'a@example.com')")
        qs = Opportunity.objects.filter(pk__in=associations.values('opportunity'))
        self.assertEqual(str(qs.values('name').query), soql)
        qs = Opportunity.objects.filter(pk__in=associations.values('opportunity_id'))
        self.assertEqual(str(qs.values('name').query), soql)
        qs = Opportunity.objects.filter(pk__in=Subquery(associations.values('opportunity')))
        self.assertEqual(str(qs.values('name').query), soql)

    def test_none_method_queryset(self):
        """Test that none() method in the queryset returns [], not error"""
        with self.lazy_assert_n_requests(0, "No database requests should be run for .none() method"):
            self.assertEqual(tuple(Contact.objects.none()), ())
            self.assertEqual(tuple(Contact.objects.all().none().all()), ())
            self.assertTrue('[]' in repr(Contact.objects.none()))
        self.lazy_check()


class TestTopologyCompiler(TestCase):
    def assertTopo(self, alias_map_items, expected):
        compiler = Contact.objects.none().query.get_compiler('salesforce')
        ret = compiler.query_topology(alias_map_items)
        self.assertEqual(ret, expected)

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_topology_compiler(self):
        # Contact.objects.all()
        # SELECT Contact.Id FROM Contact
        self.assertTopo([(None, 'Contact', None, 'Contact')], {'Contact': 'Contact'})
        # Custom.objects.all()
        # SELECT Custom__c.Id FROM Custom__c
        self.assertTopo([(None, 'Custom__c', None, 'Custom__c')], {'Custom__c': 'Custom__c'})
        # C (Id, PId) - child, P (Id) - parent
        # C.objects.filter(p__namen='xy')
        # SELECT C.Id, C.PId FROM C WHERE C.P.Name='abc'
        self.assertTopo([(None, 'C', None, 'C'), ('C', 'P', (('PId', 'Id'),), 'P')], {'C': 'C', 'P': 'C.P'})

    def test_normal2custom(self):
        # qs = Attachment.objects.filter(parent__name='abc')
        # self.assertTopo()
        # "SELECT Attachment.Id FROM Attachment WHERE Attachment.Parent.Name = 'abc'"
        self.assertTopo([('C', 'P__c', (('PId', 'Id'),), 'P__c'), (None, 'C', ((None, None),), 'C')],
                        {'P__c': 'C.P', 'C': 'C'})

    def test_custom2normal(self):
        # qs = Test.objects.filter(contact__last_name='Johnson')
        # ret = qs.query.get_compiler('salesforce').query_topology()
        # "SELECT ... FROM django_Test__c WHERE django_Test__c.Contact__r.LastName = 'Johnson'")
        self.assertTopo([(None, 'C__c', None, 'C__c'), ('C__c', 'P', (('P__c', 'Id'),), 'P')],
                        {'C__c': 'C__c', 'P': 'C__c.P__r'})

    def test_many2many(self):
        # C (Id, AId, BId) - child,  A (Id) - first parent, B (Id) - second parent
        alias_map_items = [
            (None, 'A', None, 'A'),
            ('A', 'C', (('Id', 'AId'),), 'C'),
            ('C', 'B', (('BId', 'Id'),), 'B')]
        self.assertTopo(alias_map_items, {'C': 'C', 'A': 'C.A', 'B': 'C.B'})


class QueryTest(TestCase):
    def test_where_related_in(self):
        qs = TestModel.objects.filter(contact__name__in=('a', 'b')).values('pk')
        soql = qs.query.get_compiler('salesforce').as_sql()[0]
        expected = "SELECT django_Test__c.Id FROM django_Test__c WHERE django_Test__c.Contact__r.Name IN (%s, %s)"
        self.assertEqual(soql, expected)


class SfParamsTest(TestCase):
    # type checking of this test case is currently not possible
    databases = '__all__'

    def test_params_handover_and_isolation(self):
        """Test that sp_params are propagated to the rest of the queryset chain
        but isolated from the previous part.
        """
        qs_1 = Contact.objects.all()
        qs_2 = qs_1.sf(query_all=True)
        qs_3 = qs_2.filter(first_name__gt='A')
        # a value is propagated to the next level, but not to the previous
        self.assertTrue(qs_3.query.sf_params.query_all)
        self.assertFalse(qs_1.query.sf_params.query_all)

    def test_minimal_aliases(self):
        """Test SOQL with minimal aliases without a table name of the main table"""
        # test that 'minimal_aliases' attribute is passed from qs to a salesforce compiler
        qs = Contact.objects.all()
        self.assertEqual(qs.query.get_compiler('salesforce').sf_params.minimal_aliases, False)
        self.assertEqual(qs.sf(minimal_aliases=True).query.get_compiler('salesforce').sf_params.minimal_aliases, True)

        # test that a normal SOQL is with table aliases
        qs = Contact.objects.filter(first_name='Peter').values('last_name')
        self.assertEqual(str(qs.query), "SELECT Contact.LastName FROM Contact WHERE Contact.FirstName = 'Peter'")

        # test that superfluous talbe aliases can be removed by 'minimal_aliases'
        expected_sql = "SELECT LastName FROM Contact WHERE FirstName = 'Peter'"
        qs = Contact.objects.sf(minimal_aliases=True).filter(first_name='Peter').values('last_name')
        self.assertEqual(str(qs.query), expected_sql)
        qs = Contact.objects.filter(first_name='Peter').sf(minimal_aliases=True).values('last_name')
        self.assertEqual(str(qs.query), expected_sql)
        qs = Contact.objects.filter(first_name='Peter').values('last_name').sf(minimal_aliases=True)
        self.assertEqual(str(qs.query), expected_sql)


class RegisterConversionTest(TestCase):
    @staticmethod
    def unregister_conversion(type_: Type) -> None:
        del driver.json_conversions[type_]
        del driver.sql_conversions[type_]
        if type_ in driver.subclass_conversions:
            driver.subclass_conversions.remove(type_)

    def test_arg_model_conversion(self) -> None:
        contact = Contact(last_name='test contact', pk='003001234567890AAA')

        self.assertNotIn(models.SalesforceModel, driver.json_conversions)
        self.assertNotIn(models.SalesforceModel, driver.sql_conversions)
        self.assertNotIn(models.SalesforceModel, driver.subclass_conversions)

        driver.register_conversion(models.SalesforceModel,
                                   json_conv=sobj_id,
                                   sql_conv=lambda x: driver.quoted_string_literal(sobj_id(x)),
                                   subclass=True)

        self.assertEqual(driver.arg_to_json(contact), contact.pk)
        self.assertEqual(driver.arg_to_soql(contact), contact.pk)

        self.assertIn(models.SalesforceModel, driver.json_conversions)
        self.assertIn(models.SalesforceModel, driver.sql_conversions)
        self.assertIn(models.SalesforceModel, driver.subclass_conversions)

        self.unregister_conversion(models.Model)

        self.assertNotIn(models.SalesforceModel, driver.json_conversions)
        self.assertNotIn(models.SalesforceModel, driver.sql_conversions)
        self.assertNotIn(models.SalesforceModel, driver.subclass_conversions)



"""Doctests that can be selected by 'manage.py test thismodule.doctests'"""
import doctest

doc_tests = [
    doctest.DocTestSuite('salesforce.backend.introspection'),
]


def load_tests(loader, tests, ignore):
    """Add doctests to unittests"""
    # pylint:disable=unused-argument
    tests.addTests(doc_tests)
    return tests



"""
Simple dynamic registration of admin for all yet unregistered models,
for demonstration, with respect to read only fields.
"""
from django.contrib import admin


def register_omitted_classes(models) -> None:
    """
    Register classes that don't have an own admin yet.
    Example:
        register_omitted_classes(some_app.models)
    """
    # This can be improved for fields that are only not creatable but are updateable or viceversa.
    for mdl in models.__dict__.values():
        if hasattr(mdl, '_salesforce_object') and not mdl._meta.abstract:
            attributes = {
                'readonly_fields':
                    [fld.name for fld in mdl._meta.fields if getattr(fld, 'sf_read_only', 0)]
            }
            list_display = [fld.name for fld in mdl._meta.fields if fld.column.lower() == 'name']
            if list_display:
                attributes['list_display'] = list_display
            try:
                admin.site.register(mdl,
                                    type(type(mdl).__name__ + 'Admin',
                                         (admin.ModelAdmin,),
                                         attributes)
                                    )
            except admin.sites.AlreadyRegistered:
                pass



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

import requests
from django.test import TestCase
from django.conf import settings

from salesforce import auth
from salesforce.backend.test_helpers import default_is_sf, skipUnless, sf_alias


@skipUnless(default_is_sf, "Default database should be any Salesforce.")
class OAuthTest(TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def validate_oauth(self, settings_dict):
        for key in ('access_token', 'id', 'instance_url', 'issued_at', 'signature'):
            if key not in settings_dict:
                self.fail("Missing %s key in returned oauth data." % key)           # pragma: no cover
            elif not settings_dict[key]:
                self.fail("Empty value for %s key in returned oauth data." % key)   # pragma: no cover

    def test_token_renewal(self):
        # import salesforce
        # _session=salesforce.backend.fake.base.FakeAuthSession()
        # _session.bind('default')
        _session = requests.Session()

        auth_obj = auth.SalesforcePasswordAuth(sf_alias, settings_dict=settings.DATABASES[sf_alias],
                                               _session=_session)
        auth_obj.get_auth()
        self.validate_oauth(auth.oauth_data[sf_alias])
        old_data = auth.oauth_data

        self.assertIn(sf_alias, auth.oauth_data)
        auth_obj.del_token()
        self.assertNotIn(sf_alias, auth.oauth_data)

        auth_obj.get_auth()
        self.validate_oauth(auth.oauth_data[sf_alias])

        self.assertEqual(old_data[sf_alias]['access_token'], auth.oauth_data[sf_alias]['access_token'])
        _session.close()



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Customized fields for Salesforce, especially the primary key. (like django.db.models.fields)
"""

from typing import Any, Callable, Optional, Tuple, Type, TYPE_CHECKING, Union
import warnings
from decimal import Decimal
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _
from django.db.backends.base.base import BaseDatabaseWrapper as DatabaseWrapper
from django.db.models import fields
from django.db.models import PROTECT, DO_NOTHING  # NOQA pylint:disable=unused-import
from django.db import models

from salesforce.backend import DJANGO_50_PLUS
from salesforce.defaults import DEFAULTED_ON_CREATE, DefaultedOnCreate, BaseDefault


# None of field types defined here don't need a "deconstruct" method.
# Their parameters only describe the different, but stable nature of SF standard objects.

FULL_WRITABLE = 0
NOT_UPDATEABLE = 1
NOT_CREATEABLE = 2
READ_ONLY = 3  # (NOT_UPDATEABLE & NOT_CREATEABLE)

SF_PK = getattr(settings, 'SF_PK', 'id')
if SF_PK not in ('id', 'Id'):
    raise ImproperlyConfigured("Value of settings.SF_PK must be 'id' or 'Id' or undefined.")


class SalesforceAutoField(fields.AutoField):
    """
    An AutoField that works with Salesforce primary keys.

    It is used by SalesforceModel as a custom primary key. It doesn't convert
    its value to int.
    """
    description = _("Text")

    default_error_messages = {
        'invalid': _('This value must be a valid Salesforce ID.'),
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # The parameter 'sf_read_only' is not used normally, maybe only if someone
        # added SalesforceAutoFields to the Model manually
        kwargs.pop('sf_read_only', None)
        super().__init__(*args, **kwargs)

    def to_python(self, value: Any) -> Optional[str]:
        if isinstance(value, str) or value is None:
            return value
        return str(value)

    def get_prep_value(self, value: Any) -> Any:
        return self.to_python(value)

    def contribute_to_class(self, cls: Type[models.Model], name: str, private_only: bool = False) -> None: # noqa pylint:disable=arguments-differ
        name = name if self.name is None else self.name
        # we can't require "self.auto_created==True" due to backward compatibility
        # with old migrations created before v0.6. Other conditions are enough.
        if name != SF_PK or not self.primary_key:
            raise ImproperlyConfigured(
                "SalesforceAutoField must be a primary key"
                "with the name '%s' (configurable by settings)." % SF_PK)
        if cls._meta.auto_field:
            # pylint:disable=unidiomatic-typecheck
            if not (type(self) == type(cls._meta.auto_field) and self.model._meta.abstract and  # NOQA type eq
                    cls._meta.auto_field.name == SF_PK):
                raise ImproperlyConfigured(
                    "The model %s can not have more than one AutoField, "
                    "but currently: (%s=%s, %s=%s)" % (
                        cls,
                        cls._meta.auto_field.name, cls._meta.auto_field,
                        name, self
                    )
                )
            # A model is created  that inherits fields from more abstract classes
            # with the same default SalesforceAutoField. Therefore the second should be
            # ignored.
            return
        super().contribute_to_class(cls, name, private_only=private_only)
        cls._meta.auto_field = self


class SfField(models.Field):
    """
    Add support of 'sf_read_only' and 'custom' parameters to Salesforce fields.

        sf_read_only=3 (READ_ONLY):  The field can not be specified neither on insert or update.
            e.g. LastModifiedDate (the most frequent type of read only)
        sf_read_only=1 (NOT_UPDATEABLE):  The field can be specified on insert but can not be later never modified.
            e.g. ContactId in User object (relative frequent)
        sf_read_only=2 (NOT_CREATEABLE):  The field can not be specified on insert but can be later modified.
            e.g. RecordType.IsActive or Lead.EmailBouncedReason
        sf_read_only=0:  normal writable (default)

        custom=True : Add '__c' to the column name if no db_column is defined.
    """
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.sf_read_only = kwargs.pop('sf_read_only', 0)
        self.sf_custom = kwargs.pop('custom', None)
        self.sf_namespace = ''
        self.sf_formula = kwargs.pop('sf_formula', None)
        if kwargs.get('default') is DEFAULTED_ON_CREATE:
            kwargs['default'] = DefaultedOnCreate(internal_type=self.get_internal_type())
        if 'db_default' in kwargs and not DJANGO_50_PLUS:
            del kwargs['db_default']
        super().__init__(*args, **kwargs)

    def deconstruct(self) -> Tuple[Any, Any, Any, Any]:
        name, path, args, kwargs = super().deconstruct()
        if self.name:
            _, column = self.get_attname_column()
            if '__' in column:
                kwargs.setdefault('db_column', column)
        return name, path, args, kwargs

    def get_attname_column(self) -> Tuple[str, str]:
        """
        Get the database column name automatically in most cases.
        """
        # See "A guide to Field parameters": django/db/models/fields/__init__.py
        #   * attname:   The attribute to use on the model object. This is the same as
        #                "name", except in the case of ForeignKeys, where "_id" is
        #                appended.
        #   * column:    The database column for this field. This is the same as
        #                "attname", except if db_column is specified.
        attname = self.get_attname()
        if self.db_column is not None:
            # explicit name
            column = self.db_column
        else:
            if not self.name.islower():
                # a Salesforce style name e.g. 'LastName' or 'MyCustomField'
                column = self.name
            else:
                # a Django style name like 'last_name' or 'my_custom_field'
                column = self.name.title().replace('_', '')
            # Fix custom fields
            if self.sf_custom:
                column = self.sf_namespace + column + '__c'
        return attname, column

    def contribute_to_class(self, cls: Type[models.Model], name: str, private_only: bool = False) -> None:
        super().contribute_to_class(cls, name, private_only=private_only)
        if self.sf_custom is None and hasattr(cls._meta, 'sf_custom'):
            # Only custom fields in models explicitly marked by
            # Meta custom=True are recognized automatically - for
            # backward compatibility reasons.
            self.sf_custom = cls._meta.sf_custom  # type: ignore[attr-defined]
        if self.sf_custom and '__' in cls._meta.db_table[:-3]:
            self.sf_namespace = cls._meta.db_table.split('__')[0] + '__'
        self.set_attributes_from_name(name)

# pylint:disable=unnecessary-pass,too-many-ancestors


class CharField(SfField, models.CharField):
    """CharField with sf_read_only attribute for Salesforce."""
    pass


class EmailField(SfField, models.EmailField):
    """EmailField with sf_read_only attribute for Salesforce."""
    pass


class URLField(SfField, models.URLField):
    """URLField with sf_read_only attribute for Salesforce."""
    pass


class TextField(SfField, models.TextField):
    """TextField with sf_read_only attribute for Salesforce."""
    pass


class IntegerField(SfField, models.IntegerField):
    """IntegerField with sf_read_only attribute for Salesforce."""
    pass


class BigIntegerField(SfField, models.BigIntegerField):
    """BigIntegerField with sf_read_only attribute for Salesforce."""
    # important for other database backends, e.g. in tests
    # The biggest exact value is +-(2 ** 53 -1 ), approx. 9.007E15
    pass


class SmallIntegerField(SfField, models.SmallIntegerField):
    """SmallIntegerField with sf_read_only attribute for Salesforce."""
    pass


class DecimalField(SfField, models.DecimalField):
    """
    DecimalField with sf_read_only attribute for Salesforce.

    Salesforce has only one numeric type xsd:double, but no integer.
    Even a numeric field with declared zero decimal_places can contain
    pi=3.14159265358979 in the database accidentally, but if also the value
    is integer,then it is without '.0'.
    DecimalField is the default numeric type used by itrospection inspectdb.
    """
    def to_python(self, value: Any) -> Any:
        if str(value) == '':
            return value
        ret = super().to_python(value)
        if ret is not None and self.decimal_places == 0:
            # this is because Salesforce has no numeric integer type
            if ret == int(ret):
                ret = Decimal(int(ret))
        return ret

    def from_db_value(self, value: Any, expression: Any, connection: DatabaseWrapper) -> Any:
        # pylint:disable=unused-argument
        # TODO refactor and move to the driver like in other backends
        if isinstance(value, float):
            value = str(value)
        return self.to_python(value)


class FloatField(SfField, models.FloatField):
    """FloatField for Salesforce.

    It is now really a float type not a Decimal.
    """
    def from_db_value(self, value: Any, expression: Any, connection: Any) -> Any:
        if str(value) == '' or value is None:
            return value
        return float(value)


class BooleanField(SfField, models.BooleanField):
    """BooleanField with sf_read_only attribute for Salesforce.

    Every BooleanField has a default value. It is False if default
    value checkbox is unchecked or True if checked.
    No NullBooleanField exist for Salesforce.
    """
    def __init__(self, default: Union[bool, BaseDefault] = False, **kwargs: Any) -> None:
        super().__init__(default=default, **kwargs)


class DateTimeField(SfField, models.DateTimeField):
    """DateTimeField with sf_read_only attribute for Salesforce."""


class DateField(SfField, models.DateField):
    """DateField with sf_read_only attribute for Salesforce."""

    def from_db_value(self, value: Any, expression: Any, connection: DatabaseWrapper) -> Any:
        # pylint:disable=unused-argument
        return self.to_python(value)


class TimeField(SfField, models.TimeField):
    """TimeField with sf_read_only attribute for Salesforce."""

    def from_db_value(self, value: Any, expression: Any, connection: DatabaseWrapper) -> Any:
        # pylint:disable=unused-argument
        if isinstance(value, str):
            # value from salesforce is a string "HH:MM:ss.000Z", value from other db is datetime.time().
            value = value.rstrip('Z')
        return self.to_python(value)


if TYPE_CHECKING:
    # static typing of a mixin requires an additional base, that is not necessary
    # at runtime
    _MixinTypingBase = models.ForeignObject
else:
    _MixinTypingBase = object


class SfForeignObjectMixin(SfField, _MixinTypingBase):
    def __init__(self, to: Union[Type[models.Model], str], on_delete: Callable[..., None], *args: Any, **kwargs: Any
                 ) -> None:
        # Checks parameters before call to ancestor.
        if on_delete.__name__ not in ('PROTECT', 'DO_NOTHING'):
            # The option CASCADE (currently fails) would be unsafe after a fix
            # of on_delete because Cascade delete is not usually enabled in SF
            # for safety reasons for most fields objects, namely for Owner,
            # CreatedBy etc. Some related objects are deleted automatically
            # by SF even with DO_NOTHING in Django, e.g. for
            # Campaign/CampaignMember
            warnings.warn(
                "Only foreign keys with on_delete = PROTECT or "
                "DO_NOTHING are currently supported, not %s related to %s"
                % (on_delete, to))
        super().__init__(to, on_delete, *args, **kwargs)

    def get_attname(self) -> str:
        if self.name.islower():  # pylint:disable=no-else-return
            # the same as django.db.models.fields.related.ForeignKey.get_attname
            return '%s_id' % self.name
        else:
            return '%sId' % self.name

    def get_attname_column(self) -> Tuple[str, str]:
        attname, column = super().get_attname_column()
        if self.db_column is None and not self.sf_custom:
            column += 'Id'
        return attname, column


class ForeignKey(SfForeignObjectMixin, models.ForeignKey):
    """ForeignKey with sf_read_only attribute that is acceptable by Salesforce."""


class OneToOneField(SfForeignObjectMixin, models.OneToOneField):
    """OneToOneField with sf_read_only attribute that is acceptable by Salesforce."""


class XJSONField(TextField):
    """
    Salesforce internal "complexvalue" field similar to JSON, used by SFDC for metadata,

    this field should not be used for normal data or with other database backends.
    """

    def get_internal_type(self) -> str:
        return "TextField"

    def get_prep_value(self, value: Any) -> Any:
        return value

    def to_python(self, value: Any) -> Any:
        return value


AutoField = SalesforceAutoField



"""
Minimal code to support ignored makemigrations  (like django.db.backends.*.schema)

without interaction to SF (without migrate)
"""
from django.db import NotSupportedError
from django.db.backends.base.schema import BaseDatabaseSchemaEditor

from salesforce.backend import log


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    # pylint:disable=abstract-method  # undefined: prepare_default, quote_value

    def __init__(self, connection, collect_sql=False, atomic=True):
        self.connection_orig = connection
        self.collect_sql = collect_sql
        # if self.collect_sql:
        #    self.collected_sql = []
        super().__init__(connection, collect_sql=collect_sql, atomic=atomic)

    # State-managing methods

    def __enter__(self):
        self.deferred_sql = []  # pylint:disable=attribute-defined-outside-init
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            for sql in self.deferred_sql:
                self.execute(sql)

    def execute(self, sql, params=()):
        if (sql == 'CREATE TABLE django_migrations ()'
                or sql.startswith('DROP TABLE ')) and not params:
            return
        raise NotSupportedError("Migration SchemaEditor: %r, %r" % (sql, params))

    def create_model(self, model):
        log.info("Skipped in SchemaEditor: create_model %s", model)

    # def prepare_default(self, value):    # TODO
    #     return self.quote_value(value)



import datetime
import decimal
import json
import logging
import pprint
import re
import sys
import time
import warnings
from dataclasses import dataclass
from itertools import islice
from typing import (
    Any, Callable, cast, Dict, Generic, Iterable, Iterator, List, NamedTuple, Optional,
    overload, Sequence, Tuple, Type, TypeVar, Union,
)
from urllib.parse import urlencode

import pytz
import requests
from requests.adapters import HTTPAdapter

import salesforce
from salesforce.auth import SalesforceAuth
from salesforce.dbapi.common import get_max_retries, get_thread_connections, time_statistics as time_statistics
from salesforce.dbapi.common import settings  # i.e. django.conf.settings
from salesforce.dbapi.exceptions import (  # NOQA pylint: disable=unused-import
    Error as Error, InterfaceError as InterfaceError, DatabaseError as DatabaseError, DataError as DataError,
    OperationalError as OperationalError, IntegrityError as IntegrityError, InternalError as InternalError,
    ProgrammingError as ProgrammingError, NotSupportedError as NotSupportedError,
    SalesforceError as SalesforceError, SalesforceWarning as SalesforceWarning,
    warn_sf, FakeReq, FakeResp, GenResponse)
from salesforce.dbapi.subselect import QQuery, _TRow

try:
    import beatbox as beatbox  # type: ignore[import]  # pylint: disable=unused-import,useless-import-alias
except ImportError:
    beatbox = None

log = logging.getLogger(__name__)

# --- private global constants

# Values of seconds are with 3 decimal places in SF, but they are rounded to
# whole seconds for the most of fields.
SALESFORCE_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f+0000'

# ---

request_count = 0  # global counter

ErrInfo = Tuple[Type[Exception], Exception]

ErrorHandler = Callable[['RawConnection', Optional['Cursor[Any]'], Type[BaseException], BaseException], None]


class SfSession(requests.Session):
    auth = None  # type: SalesforceAuth


class RawConnection:
    """
    parameters:
        settings_dict:  like settings.DATABASES['salesforce'] in Django
        alias:          important if the authentication is statis and should be shared for more threads
        errorhandler: function with following signature
            ``errorhandler(connection, cursor, errorclass, errorvalue)``
        use_introspection: bool
    """
    # pylint:disable=too-many-instance-attributes

    Error = Error
    InterfaceError = InterfaceError
    DatabaseError = DatabaseError
    DataError = DataError
    OperationalError = OperationalError
    IntegrityError = IntegrityError
    InternalError = InternalError
    ProgrammingError = ProgrammingError
    NotSupportedError = NotSupportedError

    def __init__(self, settings_dict: Dict[str, Any], alias: Optional[str] = None,
                 errorhandler: Optional[ErrorHandler] = None, use_introspection: Optional[bool] = None) -> None:

        thread_connections = get_thread_connections()
        if alias in thread_connections:
            del thread_connections[alias]
            raise InterfaceError(
                "Tried to open more database connections with the same alias {alias} from the same thread."
                "The previous connection has been closed now however.".format(alias=alias))
        thread_connections[alias] = self

        # private members:
        self.alias = cast(str, alias)
        self.errorhandler = errorhandler
        self.use_introspection = use_introspection if use_introspection is not None else True  # TODO implement
        self.settings_dict = settings_dict
        self.messages = []           # type: List[ErrInfo]

        self._sf_session = None      # type: Optional[SfSession]
        self._api_version = settings_dict.get('API_VERSION', salesforce.API_VERSION)  # type: str
        self.debug_verbs = []        # type: List[str]
        self.composite_type = 'sobject-collections'  # 'sobject-collections' or 'composite'

        self.sf_auth = SalesforceAuth.create_subclass_instance(db_alias=self.alias,
                                                               settings_dict=self.settings_dict)
        # The SFDC database is connected as late as possible if only tests
        # are running. Some tests don't require a connection.
        if not getattr(settings, 'SF_LAZY_CONNECT', 'test' in sys.argv):  # TODO don't use argv
            self.make_session()
        self.debug_info = {}         # type:Dict[str, Any]
        self.api_usage = ApiUsage(0, 5000)  # default before initialized by a request

    # -- public methods

    # Methods close() and commit() can be safely ignored, because everything is
    # committed automatically and the REST API is stateless.

    @property
    def api_ver(self) -> str:
        # The highest supported SFDC API version can be set by
        # settings.DATABASES['salesforce']['API_VERSION'] = 'MAX'
        # It is useful for development of 'inspectdb'
        if self._api_version == 'MAX':
            self._api_version = self.handle_api_exceptions('GET', '', api_ver='').json()[-1]['version']
        return self._api_version

    def close(self) -> None:
        del self.messages[:]
        if self._sf_session:
            self._sf_session.close()

    def commit(self) -> None:  # pylint:disable=no-self-use
        # "Database modules that do not support transactions should implement
        # this method with void functionality."
        pass

    def rollback(self) -> None:  # pylint:disable=no-self-use
        if self.settings_dict.get('OPTIONS', {}).get('WARNING_ON_ROLLBACK', False):
            log.info("Rollback is not implemented in Salesforce.")

    @overload
    def cursor(self) -> 'Cursor[Tuple[Any, ...]]': ...
    @overload  # noqa
    def cursor(self, row_type: Type[_TRow]) -> 'Cursor[_TRow]': ...

    def cursor(self, row_type=tuple):  # type: ignore[no-untyped-def]
        return Cursor(self, row_type)

    # -- private attributes

    def __enter__(self) -> 'RawConnection':
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], exc_tb: Any
                 ) -> None:
        try:
            self.close()
        except AttributeError:
            pass

    def check(self) -> None:
        if self != get_thread_connections().get(self.alias):
            raise InterfaceError("The connection has been closed previously")

    @property
    def sf_session(self) -> SfSession:
        if self._sf_session is None:
            self.make_session()
            assert self._sf_session
        return self._sf_session

    def make_session(self) -> None:
        """Authenticate and get the name of assigned SFDC data server"""
        if self._sf_session is None:
            auth_data = self.sf_auth.authenticate_and_cache()
            sf_instance_url = auth_data.get('instance_url')
            sf_session = SfSession()
            sf_session.auth = self.sf_auth  # a name for "requests" package
            if sf_instance_url and sf_instance_url not in sf_session.adapters:
                # a repeated mount to the same prefix would cause a warning about unclosed SSL socket
                sf_requests_adapter = HTTPAdapter(max_retries=get_max_retries())
                sf_session.mount(sf_instance_url, sf_requests_adapter)
            # Additional headers work, but the same are added automatically by "requests' package.
            # sf_session.header = {'accept-encoding': 'gzip, deflate', 'connection': 'keep-alive'}
            self._sf_session = sf_session

    def rest_api_url(self, *url_parts_: str, **kwargs: Any) -> str:
        """Join the URL of REST_API

        parameters:
            upl_parts:  strings that are joined to the url by "/".
                a REST url like https://na1.salesforce.com/services/data/v44.0/
                is usually added, but not if the first string starts with https://
            api_ver:  API version that should be used instead of connection.api_ver
                default. A special api_ver="" can be used to omit api version
                (for request to ask for available api versions)
            relative: If `relative` is true then the url is without domain
        Examples: self.rest_api_url("query?q=select+id+from+Organization")
                  self.rest_api_url("sobject", "Contact", id, api_ver="45.0")
                  self.rest_api_url(api_ver="")   # versions request
                  self.rest_api_url("sobject", relative=True)
                  self.rest_api_url("/services/data/v45.0")
        Output:

                  https://na1.salesforce.com/services/data/v44.0/query?q=select+id+from+Organization
                  https://na1.salesforce.com/services/data/v45.0/sobject/Contact/003DD00000000XYAAA
                  https://na1.salesforce.com/services/data
                  /services/data/v45.0
                  https://na1.salesforce.com/services/data/44.0
        """
        url_parts = list(url_parts_)
        if url_parts and re.match(r'^(?:https|mock)://', url_parts[0]):
            return '/'.join(url_parts)
        relative = kwargs.pop('relative', False)  # type: bool
        api_ver = kwargs.pop('api_ver', None)     # type: Optional[str]
        api_ver = api_ver if api_ver is not None else self.api_ver
        assert not kwargs
        if not relative:
            base = [self.sf_auth.instance_url]
        else:
            base = ['']
        if url_parts and url_parts[0].startswith('/'):
            prefix = []
            url_parts[0] = url_parts[0][1:]
        else:
            prefix = ['services/data']
            if api_ver:
                prefix += ['v{api_ver}'.format(api_ver=api_ver)]
        return '/'.join(base + prefix + url_parts)

    def handle_api_exceptions(self, method: str, *url_parts: str, **kwargs: Any) -> requests.Response:
        """Call REST API and handle exceptions
        Params:
            method:  'HEAD', 'GET', 'POST', 'PATCH' or 'DELETE'
            url_parts: like in rest_api_url() method
            api_ver:   like in rest_api_url() method
            kwargs: other parameters passed to requests.request,
                but the only notable parameter is: (... json=data)

                that works like (...
                    headers = {'Content-Type': 'application/json'},
                    data=json.dumps(data))
        """
        # The outer part - about error handler
        assert method in ('HEAD', 'GET', 'POST', 'PATCH', 'DELETE')
        cursor_context = kwargs.pop('cursor_context', None)
        errorhandler = cursor_context.errorhandler if cursor_context else self.errorhandler
        if not errorhandler:
            # nothing is caught usually and error handler not used
            return self.handle_api_exceptions_inter(method, *url_parts, **kwargs)
        try:
            return self.handle_api_exceptions_inter(method, *url_parts, **kwargs)
        except (SalesforceError, requests.exceptions.RequestException):
            exc_class, exc_value, _ = sys.exc_info()
            errorhandler(self, cursor_context, exc_class, exc_value)
            raise

    def handle_api_exceptions_inter(self, method: str, *url_parts: str, **kwargs: Any) -> requests.Response:
        """The main (middle) part - it is enough if no error occurs."""
        global request_count  # used only in single thread tests - OK # pylint:disable=global-statement
        # log.info("request %s %s", method, '/'.join(url_parts))
        api_ver = kwargs.pop('api_ver', None)
        url = self.rest_api_url(*url_parts, api_ver=api_ver)
        # The 'verify' option is about verifying TLS certificates
        kwargs_in = {'timeout': getattr(settings, 'SALESFORCE_QUERY_TIMEOUT', (4, 15)),
                     'verify': True}
        kwargs_in.update(kwargs)
        log.debug('Request API URL: %s', url)
        request_count += 1
        session = self.sf_session

        try:
            time_statistics.update_callback(url, self.ping_connection)
            response = session.request(method, url, **kwargs_in)
        except requests.exceptions.Timeout:
            raise SalesforceError("Timeout, URL=%s" % url)
        except requests.exceptions.ConnectionError as exc:
            raise SalesforceError("ConnectionError, URL=%s, %r" % (url, exc))
        if (response.status_code == 401                      # Unauthorized
                and 'json' in response.headers['content-type']
                and response.json()[0]['errorCode'] == 'INVALID_SESSION_ID'):
            # Reauthenticate and retry (expired or invalid session ID or OAuth)
            token = session.auth.reauthenticate()
            if token:
                if 'headers' in kwargs:
                    kwargs['headers'].update(Authorization='OAuth %s' % token)
                try:
                    response = session.request(method, url, **kwargs_in)
                except requests.exceptions.Timeout:
                    raise SalesforceError("Timeout, URL=%s" % url)

        if response.status_code < 400:  # OK
            # 200 "OK" (GET, POST)
            # 201 "Created" (POST)
            # 204 "No Content" (DELETE)
            # 300 ambiguous items for external ID.
            # 304 "Not Modified" (after conditional HEADER request for metadata),
            self.api_usage.update(response.headers.get('Sforce-Limit-Info'))
            return response
        # status codes docs (400, 403, 404, 405, 415, 500)
        # https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/errorcodes.htm
        self.raise_errors(response)
        return  # type: ignore[return-value]

    @staticmethod
    def raise_errors(response: GenResponse) -> None:
        """The innermost part - report errors by exceptions"""
        # Errors: 400, 403 permissions or REQUEST_LIMIT_EXCEEDED, 404, 405, 415, 500)
        # TODO extract a case ID for Salesforce support from code 500 messages

        # TODO extend or remove 'debug_verbs'. (It is disabled temporarily after writing
        #      better default messages solved the need of quied mode.)
        # verb = self.debug_verbs
        method = response.request.method
        data = None
        is_json = 'json' in response.headers.get('Content-Type', '') and response.text
        if is_json:
            data = json.loads(response.text)
        if not (isinstance(data, list) and data and 'errorCode' in data[0]):
            messages = [response.text] if is_json else []
            raise OperationalError(
                ['HTTP error "%d %s":' % (response.status_code, response.reason)]
                + messages, response, ['method+url'])

        # Other Errors are reported in the json body
        err_msg = data[0]['message']
        err_code = data[0]['errorCode']
        if response.status_code == 404:  # ResourceNotFound
            if method == 'DELETE' and err_code in ('ENTITY_IS_DELETED', 'INVALID_CROSS_REFERENCE_KEY'):
                # It was a delete command and the object is in trash bin or it is
                # completely deleted or it could be a valid Id for this sobject type.
                # Then we accept it with a warning, similarly to delete by a classic database query:
                # DELETE FROM xy WHERE id = 'something_deleted_yet'
                warn_sf([err_msg, "Object is deleted before delete or update"], response, ['method+url'])
                # TODO add a warning and add it to messages
                return None
        if err_code in ('NOT_FOUND',           # 404 e.g. invalid object type in url path or url query?q=select ...
                        'METHOD_NOT_ALLOWED',  # 405 e.g. patch instead of post
                        ):                     # both need to report the url
            raise SalesforceError([err_msg], response, ['method+url'])
        # it is good e.g for these errorCode: ('INVALID_FIELD', 'MALFORMED_QUERY', 'INVALID_FIELD_FOR_INSERT_UPDATE')
        raise SalesforceError([err_msg], response)

    def handle_api_exceptions_big(self, method: str, *url_parts: str, **kwargs: Any) -> requests.Response:
        """Call REST API with the query encapsulated into the body if the query is big"""
        assert method == 'GET'
        api_ver = kwargs.pop('api_ver', None)
        url = self.rest_api_url(*url_parts, api_ver=api_ver)
        url = re.sub(r'^\w+://[^/]+', '', url)
        data = [{'method': 'GET', 'url': url, 'referenceId': 'subrequest_0'}]
        return self.composite_request(data)

    def composite_request(self, data: List[Dict[str, Any]]) -> requests.Response:
        """Call a 'composite' request with subrequests, error handling

        A fake object for request/response is created for a subrequest in case
        of error, to be possible to use the same error hanler with a clear
        message as with an individual request.
        """
        # https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_composite_sobjects_collections.htm
        post_data = {'compositeRequest': data, 'allOrNone': True}
        resp = self.handle_api_exceptions('POST', 'composite', json=post_data)
        comp_resp = resp.json()['compositeResponse']
        is_ok = all(x['httpStatusCode'] < 400 for x in comp_resp)
        if is_ok:
            return resp

        # construct an equivalent of individual bad request/response
        bad_responses = {
            i: x for i, x in enumerate(comp_resp)
            if not (x['httpStatusCode'] == 400
                    and x['body'][0]['errorCode'] in ('PROCESSING_HALTED', 'ALL_OR_NONE_OPERATION_ROLLED_BACK'))
        }
        if len(bad_responses) != 1:
            raise InternalError("Too much or too many subrequests with an individual error")
        bad_i, bad_response = bad_responses.popitem()
        bad_request = data[bad_i]

        bad_req = FakeReq(bad_request['method'], bad_request['url'], bad_request.get('body', ''),
                          bad_request.get('httpHeaders', {}), context={bad_i: bad_request['referenceId']})

        body = [merge_dict(x, referenceId=bad_response['referenceId'])
                for x in bad_response['body']]
        bad_resp_headers = bad_response['httpHeaders'].copy()
        bad_resp_headers.update({'Content-Type': resp.headers['Content-Type']})

        bad_resp = FakeResp(bad_response['httpStatusCode'], bad_resp_headers, json.dumps(body), bad_req
                            )  # type: requests.Response # type: ignore[assignment]

        self.raise_errors(bad_resp)
        return  # type: ignore[return-value]  # TODO analyze whether this line is accessible in the case of 404 code

    @staticmethod
    def _group_results(resp_data: List[Dict[str, Any]], records: Sequence[Dict[str, Any]], all_or_none: bool
                       ) -> Tuple[List[Tuple[int, Any]], List[Tuple[int, Any, Any, str]], List[Tuple[int, Any]]]:
        x_ok, x_err, x_roll = [], [], []
        for i, x in enumerate(resp_data):
            if x['success']:
                x_ok.append((i, x))
            elif x['errors'][0]['statusCode'] in ('PROCESSING_HALTED', 'ALL_OR_NONE_OPERATION_ROLLED_BACK'):
                x_roll.append((i, x))
            else:
                if isinstance(records[i], dict):
                    bad_id = [v for k, v in records[i].items() if k.lower() == 'id']
                    x_err.append((i, x['errors'], records[i]['attributes']['type'], bad_id[0] if bad_id else None))
                else:
                    x_err.append((i, x['errors'], 'unknown_type', records[i]))
        if all_or_none:
            # more errors can be reported even with all_or_none, but sometimes only the first concrete error
            assert not x_err and not x_roll or not x_ok and len(x_err) > 0
        else:
            assert not x_roll
        return x_ok, x_err, x_roll

    def sobject_collections_request(self,
                                    method: str,
                                    records: Sequence[Dict[str, Any]],
                                    all_or_none: bool = True
                                    ) -> List[str]:
        # https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_composite_composite.htm
        # pylint:disable=too-many-locals
        assert method in ('GET', 'POST', 'PATCH', 'DELETE')
        if method == 'DELETE':
            assert all(isinstance(x, str) for x in records)
            ids = cast(Sequence[str], records)
            params = dict(ids=','.join(ids), allOrNone=str(bool(all_or_none)).lower())
            resp = self.handle_api_exceptions(method, 'composite/sobjects', params=params)
        else:
            assert all(isinstance(x, dict) for x in records)
            if method in ('POST', 'PATCH'):
                records = [merge_dict(x, attributes={'type': x['type_']}) for x in records]
                for x in records:
                    x.pop('type_')
                post_data = {'records': records, 'allOrNone': all_or_none}
            else:
                raise NotSupportedError("Method {} not implemended".format(method))

            resp = self.handle_api_exceptions(method, 'composite/sobjects', json=post_data)
        resp_data = resp.json()

        x_ok, x_err, x_roll = self._group_results(resp_data, records, all_or_none)  # pylint:disable=unused-variable
        is_ok = not x_err
        if is_ok:
            return [x['id'] for i, x in x_ok]  # for .lastrowid

        width_type = max(len(type_) for i, errs, type_, id_ in x_err)
        width_type = max(width_type, len('sobject'))
        messages = [
            '(see details below)',
            '',
            'Error Summary: errors={}, rollback/cancel={}, success={}'.format(len(x_err), len(x_roll), len(x_ok)),
            'index {} sobject{:{width}s}error_info'.format(
                ('ID' + 16 * ' ' if x_err[0][3] else ''), '', width=(width_type + 2 - len('sobject')))
        ]
        for i, errs, type_, id_ in x_err:
            field_info = 'FIELDS: {}'.format(errs[0]['fields']) if errs[0].get('fields') else ''
            msg = '{:5d} {} {:{width_type}s}  {}: {} {}'.format(
                i, id_ or '', type_, errs[0]['statusCode'], errs[0]['message'], field_info,
                width_type=width_type)
            messages.append(msg)
        raise SalesforceError(messages)

    def ping_connection(self, timeout: float = 1.0) -> float:
        """Fast check the connection by an unimportant request

        It is useful after a longer inactivity if a connection could
        have been incorrectly terminated and can cause a timeout. This
        simple command will recreate the connection after a short timeout
        1 second, if necessary, while normal commands use a longer timeout,
        typically 30 sec.
        Returns the duration if the command succeded.
        """
        t_0 = time.time()
        try:
            self.handle_api_exceptions('GET', '', api_ver='', timeout=timeout)
        except (requests.exceptions.RequestException, SalesforceError):
            pass
        return round(time.time() - t_0, 3)


Connection = RawConnection


# DB API function
def connect(**params: Any) -> Connection:
    return Connection(**params)


def get_connection(alias: str, **params: Any) -> Connection:
    if alias not in get_thread_connections():
        connect(alias=alias, **params)
    return cast(Connection, get_thread_connections()[alias])


class Cursor(Generic[_TRow]):
    """Cursor (local part is simple, remote part is stateless for small result sets)

    From SDFC documentation (combined from SOQL, REST, APEX)
    Query results are returned from SFDC by blocks of 2000 rows, by default.
    It is guaranted that the size of result block is not smaller than 200 rows,
    and not longer than 2000 rows.
    - A maximum of 10 open locators per user are available. The oldest locator
      (cursor) is deleted then by SFDC.
    - Query locators older than 15 minutes are deleted. (usually after much longer time)
    - The query locator works without transactions
    and it should be expected that the data after 2000 rows could be affected by
    later updates and may not match the `where` condition.

    The local part of the cursor is only an index into the last block of 2000
    rows. The number of local cursors is unrestricted.

    Queries with binary fields Base64 (e.g. attachments) are returned by one row.
    the best it to get more rows without binary fields and then to query
    for complete rows, one by one.
    """

    # pylint:disable=too-many-instance-attributes,too-many-public-methods
    def __init__(self, connection: Connection, row_type: Optional[Type[_TRow]] = None) -> None:
        # DB API attributes (public, ordered by documentation PEP 249)
        self.description = None           # type: Optional[List[Tuple[Any, ...]]]
        self.rowcount = -1  # set to non-negative by SELECT INSERT UPDATE DELETE
        self.arraysize = 1  # writable, but ignored finally
        # db api extensions
        self.rownumber = None             # type: Optional[int]  # cursor position index
        self._connection = connection     # type: Connection
        self.messages = []                # type: List[ErrInfo]
        self.lastrowid = None  # TODO to be used for INSERT id, but insert is not implemented by cursor
        self.errorhandler = connection.errorhandler
        # private
        assert row_type in (tuple, list, dict, None)
        if row_type is None:
            self.row_type = tuple         # type: Union[Type[Dict[str, Any]], Type[Tuple[Any, ...]], Type[List[Any]]]
        else:
            self.row_type = row_type
        self._chunk = []                  # type: List[Dict[str, Any]]  # it is in the native JSON format
        self._chunk_offset = None         # type: Optional[int]
        self._next_records_url = None     # type: Optional[str]
        self.handle = None                # type: Optional[str]
        self.qquery = None                # type: Optional[QQuery]
        self._raw_iterator = None         # type: Optional[Iterator[Dict[str, Any]]]
        self._iter = not_executed_yet()   # type: Iterator[_TRow]
        self.closed = False

    # -- DB API methods

    # .callproc(...)  noit implemented

    def close(self) -> None:
        self._clean()
        self.closed = True

    def execute(self, soql: str, parameters: Optional[Iterable[Any]] = None, query_all: bool = False,
                tooling_api: bool = False) -> None:
        self._clean()
        parameters = parameters or []
        if 'use_debug_info' in self.connection.debug_verbs:
            processed_soql = str(soql) % tuple(arg_to_soql(x) for x in parameters)
            self.connection.debug_info['soql'] = (soql, parameters, processed_soql)
        sqltype = soql.split(None, 1)[0].upper()
        if sqltype == 'SELECT':
            self.execute_select(soql, parameters, query_all=query_all, tooling_api=tooling_api)
        elif sqltype == 'EXPLAIN':
            assert not tooling_api
            self.execute_explain(soql, parameters, query_all=query_all)
        else:
            # INSERT UPDATE DELETE
            raise ProgrammingError("Unexpected command '{}'".format(sqltype))

    def executemany(self, operation: str, seq_of_parameters: Iterable[Iterable[Any]]) -> None:
        self._clean()
        for param in seq_of_parameters:
            self.execute(operation, param)

    def fetchone(self) -> Optional[_TRow]:
        self._check_data()
        return next(self, None)

    def fetchmany(self, size: Optional[int] = None) -> List[_TRow]:
        self._check_data()
        if size is None:
            size = self.arraysize
        return list(islice(self, size))

    def fetchall(self) -> List[_TRow]:
        self._check_data()
        return list(self)

    def scroll(self, value: int, mode: str = 'relative') -> None:
        # TODO It is a beta based on an undocumented information
        # The undocumented structure of 'nextRecordsUrl' is
        #     'query/{query_id}-{offset}'  e.g.
        #     '/services/data/v44.0/query/01gM000000zr0N0IAI-3012'
        warnings.warn("cursor.scroll is based on an undocumented Salesforce info. "
                      "Long jump in a big query may be unsupported.")
        assert mode in ('absolute', 'relative')
        # The possible offset range is from 0 to resp.json()['totalSize']
        # An exception IndexError should be raised if the offset is invalid [PEP 429]
        # but this will simply pass the original exception from SFDC:
        # "SalesforceError: INVALID_QUERY_LOCATOR"

        assert self._chunk_offset is not None and self.rownumber is not None
        new_offset = int(value) + (0 if mode == 'absolute' else self.rownumber)
        if not self._chunk_offset <= new_offset < self._chunk_offset + len(self._chunk):
            url = '{}-{}'.format(self.handle, new_offset)
            self.query_more(url)
            self._chunk_offset = new_offset
        self._raw_iterator = iter(self._chunk)
        if new_offset != self._chunk_offset:
            rel_offs = new_offset - self._chunk_offset
            next(islice(self._raw_iterator, rel_offs, rel_offs), None)
        self.rownumber = new_offset
        self._iter = iter(self._gen())

    # .nextset()  not implemented

    def setinputsizes(self, sizes: Any) -> None:
        pass  # this method is allowed to do nothing

    def setoutputsize(self, size: int, column: Any = None) -> None:
        pass  # this method is allowed to do nothing

    def __next__(self) -> _TRow:
        return self._iter.__next__()

    def commit(self) -> None:
        self._connection.commit()

    def rollback(self) -> None:
        self._connection.rollback()

    # -- private methods

    def __enter__(self) -> 'Cursor[_TRow]':
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], exc_tb: Any
                 ) -> None:
        try:
            self.close()
        except AttributeError:
            pass

    @property
    def connection(self) -> Connection:
        self.check()
        return self._connection

    def check(self) -> None:
        self._connection.check()
        if self.closed:
            raise InterfaceError("Cursor is closed")

    def __iter__(self) -> 'Cursor[_TRow]':
        return self

    def _gen(self) -> Iterator[_TRow]:
        assert self._chunk_offset is not None and self.rownumber is not None
        assert self.qquery

        while True:
            self._raw_iterator = iter(self._chunk)
            for row in self.qquery.parse_rest_response(self._raw_iterator, self.rowcount,
                                                       row_type=self.row_type):
                yield cast(_TRow, row)
                self.rownumber += 1
            if not self._next_records_url:
                break
            new_offset = self._chunk_offset + len(self._chunk)
            self.query_more(self._next_records_url)
            self._chunk_offset = new_offset

    def execute_select(self, soql: str, parameters: Iterable[Any], query_all: bool = False,
                       tooling_api: bool = False) -> None:
        processed_sql = str(soql) % tuple(arg_to_soql(x) for x in parameters)
        service = '' if not tooling_api else 'tooling/'
        service += 'query' if not query_all else 'queryAll'

        self.qquery = qquery = QQuery(soql)
        # TODO better description
        self.description = [(alias, None, None, None, name) for alias, name in
                            zip(qquery.aliases, qquery.fields)]

        url_part = '/?'.join((service, urlencode(dict(q=processed_sql))))
        self.query_more(url_part)
        self._chunk_offset = 0
        self.rownumber = 0
        if self._next_records_url:
            self.handle = self._next_records_url.split('-')[0]
        self._iter = iter(self._gen())

    def execute_explain(self, soql: str, parameters: Iterable[Any], query_all: bool = False) -> None:
        # https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_query_explain.htm
        self._clean()
        assert soql.startswith('EXPLAIN SELECT')
        soql = soql.split(' ', 1)[1]
        processed_sql = str(soql) % tuple(arg_to_soql(x) for x in parameters)
        service = 'query' if not query_all else 'queryAll'

        self.qquery = QQuery(soql)
        self.description = [('detail', None, None, None, 'detail')]
        url_part = '/?'.join((service, urlencode(dict(explain=processed_sql))))
        ret = self.handle_api_exceptions('GET', url_part)

        self._chunk = [{'explain': x} for x in pprint.pformat(ret.json(), indent=1, width=100).split('\n')]
        self._chunk_offset = 0
        self.rownumber = 0
        self._iter = iter(self._gen())

    def query_more(self, nextRecordsUrl: str) -> None:
        self._check()
        if len(nextRecordsUrl) < 15500:
            ret = self.handle_api_exceptions('GET', nextRecordsUrl).json()
        else:
            ret = self.connection.handle_api_exceptions_big('GET', nextRecordsUrl).json()
            ret = ret['compositeResponse'][0]['body']
        self.rowcount = ret['totalSize']  # may be more accurate than the initial approximate value
        self._chunk = ret['records']
        self._next_records_url = ret.get('nextRecordsUrl')

    def _check(self) -> None:
        if not self.connection:
            raise InterfaceError("Cursor Closed")

    def _check_data(self) -> None:
        if not self._iter:
            raise ProgrammingError('No previous .execute("select...") before .fetch...()')

    def _clean(self) -> None:
        self.description = None
        self.rowcount = -1
        self.rownumber = None
        del self.messages[:]
        self.lastrowid = None
        self._next_records_url = None
        self._chunk = []
        self._chunk_offset = None
        self.handle = None
        self.qquery = None
        self._raw_iterator = None
        self._iter = not_executed_yet()
        self._check()

    def handle_api_exceptions(self, method: str, *url_parts: str, **kwargs: Any) -> requests.Response:
        return self.connection.handle_api_exceptions(method, *url_parts, cursor_context=self, **kwargs)

    # --- custom extension methods

    def urls_request(self) -> Any:  # Dict[str, Any]
        """Empty REST API request is useful after long inactivity before POST.

        It ensures that the token will remain valid for at least half life time
        of the new token. Otherwise it would be an awkward doubt if a timeout on
        a lost connection is possible together with token expire in a post
        request (insert).
        """
        ret = self.handle_api_exceptions('GET', '')
        return ret.json()

    def id_request(self) -> Any:  # Dict[str, Any]
        """The Force.com Identity Service (return type dict of str)"""
        # https://developer.salesforce.com/page/Digging_Deeper_into_OAuth_2.0_at_Salesforce.com?language=en&language=en#The_Force.com_Identity_Service
        oauth = self.connection.sf_session.auth.get_auth()
        if 'id' in oauth:
            url = oauth['id']
        else:
            # dynamic auth without 'id' parameter
            url = self.urls_request()['identity']
        ret = self.handle_api_exceptions('GET', url)  # TODO
        return ret.json()

    def versions_request(self) -> Any:  # List[Dict[str, str]]
        """List Available REST API Versions"""
        return self.handle_api_exceptions('GET', '', api_ver='').json()


# ---

CursorDescription = NamedTuple(
    'CursorDescription', [
        ('name', str),          # The first two items are mandatory. (name, type)
        ('type_code', Any),
        ('display_size', bool),
        ('internal_size', int),
        ('precision', int),
        ('scale', int),
        ('null_ok', bool),
        ('default', Any),
        ('params', dict),  # type: ignore[type-arg]
    ])
CursorDescription.__new__.__defaults__ = 7 * (None,)  # type: ignore[attr-defined]


def standard_errorhandler(connection: Connection, cursor: Optional[Cursor[Any]], errorclass: Type[Exception],
                          errorvalue: Exception) -> None:
    if cursor:
        cursor.messages.append((errorclass, errorvalue))
    else:
        connection.messages.append((errorclass, errorvalue))


def verbose_error_handler(connection: Connection, cursor: Optional[Cursor[Any]],  # pylint:disable=unused-argument
                          errorclass: Type[Exception], errorvalue: Exception,  # pylint:disable=unused-argument
                          ) -> None:
    pprint.pprint(errorvalue.__dict__)


@dataclass
class ApiUsage:
    # see https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/headers_api_usage.htm
    api_usage: int  # API requests used per last 24 hours
    api_limit: int  # API requests limit pe 24 hours

    def update(self, sforce_limit_info: Optional[str]) -> None:
        # example: .update('api-usage=692/5000000')
        if sforce_limit_info:
            key, val = sforce_limit_info.split('=')
            assert key == 'api-usage'
            api_usage_s, api_limit_s = val.split('/')
            self.api_usage = int(api_usage_s)
            self.api_limit = int(api_limit_s)


# --- private


def not_executed_yet() -> Iterator[_TRow]:
    raise Connection.InterfaceError("called fetch...() before execute()")
    yield  # pylint:disable=unreachable


def signalize_extensions() -> None:
    """DB API 2.0 extension are reported by warnings at run-time."""
    warnings.warn("DB-API extension cursor.rownumber used", SalesforceWarning)
    warnings.warn("DB-API extension connection.<exception> used", SalesforceWarning)  # TODO
    warnings.warn("DB-API extension cursor.connection used", SalesforceWarning)
    # not implemented DB-API extension cursor.scroll(, SalesforceWarning)
    warnings.warn("DB-API extension cursor.messages used", SalesforceWarning)
    warnings.warn("DB-API extension connection.messages used", SalesforceWarning)
    warnings.warn("DB-API extension cursor.next(, SalesforceWarning) used")
    warnings.warn("DB-API extension cursor.__iter__(, SalesforceWarning) used")
    warnings.warn("DB-API extension cursor.lastrowid used", SalesforceWarning)
    warnings.warn("DB-API extension .errorhandler used", SalesforceWarning)


# ----

# basic conversions

T = TypeVar('T')
ConversionSqlFunc = Callable[[T], Union[str, float]]
ConversionJsonFunc = Callable[[T], Optional[Union[str, float]]]


def register_conversion(type_: Type[T],
                        json_conv: ConversionJsonFunc[Any],
                        sql_conv: Optional[ConversionSqlFunc[Any]] = None,
                        subclass: bool = False
                        ) -> None:
    json_conversions[type_] = json_conv
    sql_conversions[type_] = cast(ConversionSqlFunc[Any], sql_conv or json_conv)
    if subclass and type_ not in subclass_conversions:
        subclass_conversions.append(type_)


def quoted_string_literal(txt: str) -> str:
    """
    SOQL requires single quotes to be escaped.
    http://www.salesforce.com/us/developer/docs/soql_sosl/Content/sforce_api_calls_soql_select_quotedstringescapes.htm
    """
    try:
        return "'%s'" % (txt.replace("\\", "\\\\").replace("'", "\\'"),)
    except TypeError:
        raise NotSupportedError("Cannot quote %r objects: %r" % (type(txt), txt))


def date_literal(dat: datetime.datetime) -> str:
    if not dat.tzinfo:
        tz = pytz.timezone(settings.TIME_ZONE)
        dat = tz.localize(dat, is_dst=bool(time.daylight))
    # Format of `%z` is "+HHMM"
    tzname = datetime.datetime.strftime(dat, "%z")
    return datetime.datetime.strftime(dat, "%Y-%m-%dT%H:%M:%S.000") + tzname


def arg_to_soql(arg: Any) -> Union[str, float]:
    """
    Perform necessary SOQL quoting on the arg.
    """
    conversion = sql_conversions.get(type(arg))
    if conversion:
        return conversion(arg)
    for type_ in subclass_conversions:
        if isinstance(arg, type_):
            return sql_conversions[type_](arg)
    return sql_conversions[str](arg)


def arg_to_json(arg: Any) -> Optional[Union[str, float]]:
    """
    Perform necessary JSON conversion on the arg.
    """
    conversion = json_conversions.get(type(arg))
    if conversion:
        return conversion(arg)
    for type_ in subclass_conversions:
        if isinstance(arg, type_):
            return json_conversions[type_](arg)
    return json_conversions[str](arg)


# supported types converted from Python to SFDC

# conversion before conversion to json (for Insert and Update commands)
json_conversions = {}  # type: Dict[Type[Any], ConversionJsonFunc[Any]]

# conversion before formating a SOQL (for Select commands)
sql_conversions = {}  # type: Dict[Type[Any], ConversionSqlFunc[Any]]

subclass_conversions = []  # type: List[type]

register_conversion(int,             json_conv=str)
register_conversion(float,           json_conv=lambda o: '%.15g' % o)
register_conversion(type(None),      json_conv=lambda s: None,          sql_conv=lambda s: 'NULL')
register_conversion(str,             json_conv=lambda o: cast(str, o),  sql_conv=quoted_string_literal)  # default
register_conversion(bool,            json_conv=lambda s: str(s).lower())
register_conversion(datetime.date,   json_conv=lambda d: datetime.date.strftime(d, "%Y-%m-%d"))
register_conversion(datetime.datetime, json_conv=date_literal)
register_conversion(datetime.time,   json_conv=lambda d: datetime.time.strftime(d, "%H:%M:%S.%fZ"))
register_conversion(decimal.Decimal, json_conv=float, subclass=True)
# the type models.Model is registered from backend, because it is a Django type


def merge_dict(dict_1: Dict[Any, Any], **kw: Any) -> Dict[Any, Any]:
    """Merge a dict with some keys and values (or another dict)."""
    return {**dict_1, **kw}



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Salesforce object manager. (like django.db.models.manager)

Use a custom QuerySet to generate SOQL queries and results.

This module requires a customized package django-stubs (django-salesforce-stubs)
"""

from typing import Generic, Optional, TypeVar
from django.db.models import manager, Model
from django.db.models.query import QuerySet  # pylint:disable=unused-import

from salesforce import router
from salesforce.backend import query

_T = TypeVar("_T", bound=Model, covariant=True)


class SalesforceManager(manager.Manager, Generic[_T]):

    def get_queryset(self, _alias: Optional[str] = None) -> 'QuerySet[_T]':
        """
        Returns a QuerySet which access remote SF objects.
        """
        alias_is_sf = _alias and router.is_sf_database(_alias)
        is_extended_model = getattr(self.model, '_salesforce_object', '') == 'extended'
        assert self.model is not None
        if router.is_sf_database(self.db) or alias_is_sf or is_extended_model:
            return query.SalesforceQuerySet(self.model, using=self.db)
        return super().get_queryset()

    # def raw(self, raw_query, params=None, translations=None):
    #     if router.is_sf_database(self.db):
    #         q = models_sql_query.SalesforceRawQuery(raw_query, self.db, params)
    #         return query.SalesforceRawQuerySet(raw_query=raw_query, model=self.model, query=q,
    #                                            params=params, using=self.db)
    #     return super().raw(raw_query, params=params, translations=translations)

    # methods of SalesforceQuerySet on a SalesfroceManager

    def query_all(self) -> 'query.SalesforceQuerySet[_T]':  # type: ignore[override] # noqa
        qs = self.get_queryset()
        assert isinstance(qs, query.SalesforceQuerySet)
        ret = qs.query_all()
        return ret

    def sf(self,
           query_all: Optional[bool] = None,
           all_or_none: Optional[bool] = None,
           edge_updates: Optional[bool] = None,
           minimal_aliases: Optional[bool] = None) -> 'query.SalesforceQuerySet[_T]':
        # not dry, but explicit due to preferring type check of user code
        qs = self.get_queryset()
        assert isinstance(qs, query.SalesforceQuerySet)
        return qs.sf(
            query_all=query_all,
            all_or_none=all_or_none,
            edge_updates=edge_updates,
            minimal_aliases=minimal_aliases,
        )



import re
import threading
import time
from typing import Any, cast, Callable, Dict, Optional

# Dependencies of salesforce.dbapi on Django are minimalized.
# The biggest challenges are django.conf.settings, django.db.connections and django.test.

try:
    settings = None  # type: Any
    from django.conf import settings
except ImportError:
    # mock for some tests without a real Django
    import importlib
    import os

    settings = importlib.import_module(os.environ['DJANGO_SETTINGS_MODULE'])

    if not getattr(settings, 'SF_CAN_RUN_WITHOUT_DJANGO', False):
        raise


def get_max_retries() -> int:
    """Get the maximal number of requests retries

    The maximal number of retries for timeouts in requests to Force.com API.
    Can be set dynamically
    None: use defaults from settings.REQUESTS_MAX_RETRIES (default 1)
    0: no retry
    1: one retry
    """
    return getattr(settings, 'REQUESTS_MAX_RETRIES', 1)  # type: ignore[no-any-return] # noqa


def get_thread_connections() -> Dict[Optional[str], Any]:
    if not hasattr(thread_loc, 'connections'):
        thread_loc.connections = {}
    return cast(Dict[Optional[str], Any], thread_loc.connections)


class TimeStatistics:

    def __init__(self, expiration: float = 300) -> None:
        self.expiration = expiration
        self.data = {}  # type: Dict[str, float]

    def update_callback(self, url: str, callback: Optional[Callable[[], Any]] = None) -> None:
        """Update the statistics for the domain"""
        domain = self.domain(url)
        last_req = self.data.get(domain, 0)
        t_new = time.time()
        do_call = (t_new - last_req > self.expiration)
        self.data[domain] = t_new
        if do_call and callback:
            callback()

    @staticmethod
    def domain(url: str) -> str:
        match = re.match(r'^(?:https|mock)://([^/]*)/?', url)
        assert match, "HOST must be including the protocol and :// like 'https://login.salesforce.com'"
        return match.groups()[0]


time_statistics = TimeStatistics(300)
thread_loc = threading.local()



import sys
import traceback
import unittest
from contextlib import contextmanager
from unittest import expectedFailure
from unittest import mock  # pylint:disable=unused-import  # NOQA
from typing import Any, Callable, List, Iterator, Optional, Tuple, Type, TYPE_CHECKING

from salesforce.dbapi import driver
from salesforce.dbapi.common import get_thread_connections
TestMethod = Callable[..., None]


def expectedFailureIf(condition: bool) -> Callable[[TestMethod], TestMethod]:
    """Conditional 'expectedFailure' decorator for TestCase"""
    if condition:
        return expectedFailure
    return lambda func: func


class QuietSalesforceErrors:
    """Context manager that helps expected SalesforceErrors to be not logged too verbose.

    It works on the default Salesforce connection. It can be nested.
    """
    def __init__(self, alias: Optional[str], verbs: Optional[List[str]] = None):
        self.connection = get_thread_connections()[alias or 'salesforce']
        self.verbs = verbs or []
        self.save_debug_verbs = None

    def __enter__(self) -> 'QuietSalesforceErrors':
        self.save_debug_verbs = self.connection.debug_verbs
        self.connection.debug_verbs = self.verbs
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], exc_tb: Any
                 ) -> None:
        try:
            self.connection.debug_verbs = self.save_debug_verbs
        except AttributeError:
            pass


class PatchedSfConnection(QuietSalesforceErrors):
    """A connection with a temporarily modified setting"""
    # this is only a mnemonic name for other usage of the parent class


if TYPE_CHECKING:
    BaseLazyTestMixin = unittest.TestCase
else:
    BaseLazyTestMixin = object


class LazyTestMixin(BaseLazyTestMixin):
    """Report less important asserts lazily, only if no more important failure occurs.

    Only the first lazy_assert(...) failure is reported at the end of test at "lazy_check()"
    and only if no more important failure occurs between lazy_assert and lazy_check.

        self.lazy_assert(expr...)
        ...
        self.assertEqual(expr...)
        ...
        self.lazy_assert(expr...)
        ...
        self.assertEqual(expr...)
        ...
        self.assert_check()
    """
    class LazyAssertionError(AssertionError):
        pass

    def setUp(self) -> None:
        self.lazy_failure = None  # type: Optional[Tuple[Type[BaseException], BaseException, str]]

    def _lazy_assert(self, assert_method: TestMethod, *args: Any, **kwargs: Any) -> None:
        def stack_of_test_method(raw_stack: List[Any]) -> List[str]:
            stack = traceback.format_list(raw_stack)
            parent_frame = max((i for i, x in enumerate(stack) if x.endswith(' testMethod()\n')), default=0)
            if not parent_frame:
                # this is for Python >= 3.8
                parent_frame = 1 + max(
                    (i for i, x in enumerate(stack) if x.endswith('_callTestMethod(testMethod)\n')), default=-1)
            return stack[1 + parent_frame:-skip_frames]

        if not getattr(self, 'lazy_failure', None):
            skip_frames = kwargs.pop('skip_frames', 2)
            # the name is like in unittests due to more readable test tracebacks
            lazyAssertMethod = assert_method
            try:
                lazyAssertMethod(*args, **kwargs)
            except self.failureException:
                exc_type, exc_value, _ = sys.exc_info()
                stack = stack_of_test_method(traceback.extract_stack())
                assert exc_type and exc_value
                self.lazy_failure = exc_type, exc_value, ''.join(stack).strip()

    def lazy_check(self) -> None:
        """Check all previous `lazy_assert_*`"""
        lazy_failure = getattr(self, 'lazy_failure', None)
        if lazy_failure and get_thread_connections():
            # check it only if any Salesforce connection is initialized
            exc_type, original_exception, traceback_part_text = lazy_failure
            msg = original_exception.args[0]
            original_exception = self.LazyAssertionError(
                '\n'.join((
                    '\nUse the first frame of this real traceback and ignore the previous lazy traceback:',
                    traceback_part_text,
                    '{}: {}'.format(exc_type.__name__, msg),
                )),
                *original_exception.args[1:]
            )
            raise original_exception

    def lazyAssertTrue(self, expr: Any, msg: Optional[str] = None) -> None:
        self._lazy_assert(self.assertTrue, expr, msg=msg)

    def lazyAssertEqual(self, expr_1: Any, expr_2: Any, msg: Optional[str] = None) -> None:
        self._lazy_assert(self.assertEqual, expr_1, expr_2, msg=msg)

    @contextmanager
    def lazy_assert_n_requests(self, expected_requests: int, msg: Optional[str] = None) -> Iterator[None]:
        """Assert that the inner block requires expected_requests, evaluated lazily."""
        request_count_0 = driver.request_count
        try:
            yield None
        finally:
            request_count_1 = driver.request_count
            msg = (msg + '\n') if msg else ''
            msg += ('expected requests != real requests;  checked by:\n'
                    '    with self.lazy_assert_n_requests({}):'.format(expected_requests))
            self.lazyAssertEqual(expected_requests, request_count_1 - request_count_0, msg=msg)



"""This file is useful only if 'salesforce' is a duplicit name in Django registry

then put a string 'salesforce.apps.SalesforceDb' instead of simple 'salesforce'
"""
from django.apps import AppConfig


class SalesforceDb(AppConfig):
    name = 'salesforce'
    label = 'salesforce_db'



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Django models for accessing Salesforce objects. (like django.db.models)

The Salesforce database is somewhat un-UNIXy or non-Pythonic, in that
column names are all in CamelCase. No attempt is made to work around this
issue, but normal use of `db_column` and `db_table` parameters should work.
"""

from inspect import isclass
from typing import Any, Dict, Generic, Iterable, TYPE_CHECKING, TypeVar
import logging
import re
import types
import warnings

from django.db import models
from django.db.models.base import ModelBase
# Only these two `on_delete` options are currently supported
from django.db.models import PROTECT, DO_NOTHING  # NOQA pylint:disable=unused-wildcard-import,wildcard-import
# from django.db.models import CASCADE, PROTECT, SET_NULL, SET, DO_NOTHING

from salesforce.defaults import DefaultedOnCreate, DEFAULTED_ON_CREATE
from salesforce.fields import (
    SalesforceAutoField as SalesforceAutoField, SF_PK, SfField, ForeignKey as ForeignKey)
from salesforce.fields import (NOT_UPDATEABLE as NOT_UPDATEABLE, NOT_CREATEABLE as NOT_CREATEABLE,
                               READ_ONLY as READ_ONLY)
from salesforce.fields import (  # noqa pylint:disable=useless-import-alias  # for other modules, but unused here
    AutoField as AutoField, BigIntegerField as BigIntegerField, BooleanField as BooleanField,
    CharField as CharField, DateField as DateField, DateTimeField as DateTimeField,
    DecimalField as DecimalField, EmailField as EmailField, FloatField as FloatField,
    IntegerField as IntegerField, OneToOneField as OneToOneField, SmallIntegerField as SmallIntegerField,
    TextField as TextField, TimeField as TimeField, URLField as URLField, XJSONField as XJSONField,
)
from salesforce.fields import *  # NOQA pylint:disable=unused-wildcard-import,wildcard-import
from salesforce.backend.indep import LazyField
if not TYPE_CHECKING:
    # the SalesforceManager and the module salesforce.backend.manager whould be imported
    # at run-time because it breaks the generic mechanism in `django-stubs` v1.5,
    # that assigns Manager[_T] (a Django manager for the model _T).
    from salesforce.backend import manager

__all__ = ('SalesforceModel', 'Model', 'DEFAULTED_ON_CREATE', 'PROTECT', 'DO_NOTHING', 'SF_PK', 'SfField',
           'NOT_UPDATEABLE', 'NOT_CREATEABLE', 'READ_ONLY', 'DefaultedOnCreate')

log = logging.getLogger(__name__)

_T = TypeVar('_T', bound="models.Model", covariant=True)


class SalesforceModelBase(ModelBase):
    """
    This is a sub-metaclass of the normal Django ModelBase.

    This metaclass overrides the default table-guessing behavior of Django
    and replaces it with code that defaults to the model name.
    """
    # pylint: disable=no-value-for-parameter
    def __new__(cls, name, bases, attrs, **kwargs):  # type: ignore [no-untyped-def]
        attr_meta = attrs.get('Meta', None)
        supplied_db_table = getattr(attr_meta, 'db_table', None)
        if getattr(attr_meta, 'dynamic_field_patterns', None):
            pattern_module, dynamic_field_patterns = getattr(attr_meta, 'dynamic_field_patterns')
            make_dynamic_fields(pattern_module, dynamic_field_patterns, attrs)
            delattr(attr_meta, 'dynamic_field_patterns')
        result = super(SalesforceModelBase, cls     # type: ignore [call-overload]
                       ).__new__(cls, name, bases, attrs, **kwargs)
        if models.Model not in bases and supplied_db_table is None:
            result._meta.db_table = result._meta.concrete_model._meta.object_name
            result._meta.original_attrs['db_table'] = result._meta.db_table
        return result

    def add_to_class(cls, name: str, value: Any) -> Any:
        # pylint:disable=protected-access
        if name == '_meta':
            sf_custom = False
            sf_tooling_api_model = False
            if hasattr(value.meta, 'custom'):
                sf_custom = value.meta.custom
                delattr(value.meta, 'custom')
            if hasattr(value.meta, 'sf_tooling_api_model'):
                sf_tooling_api_model = value.meta.sf_tooling_api_model
                delattr(value.meta, 'sf_tooling_api_model')
            super(SalesforceModelBase, cls).add_to_class(name, value)  # type: ignore[misc]
            setattr(cls._meta, 'sf_custom', sf_custom)  # type: ignore[attr-defined]
            setattr(cls._meta, 'sf_tooling_api_model', sf_tooling_api_model)  # type: ignore[attr-defined]
        else:
            if type(value) is models.manager.Manager:  # pylint:disable=unidiomatic-typecheck
                # this is for better migrations because: obj._constructor_args = (args, kwargs)
                _constructor_args = value._constructor_args
                value = models.manager.Manager()
                # value = manager.SalesforceManager()
                value._constructor_args = _constructor_args

            super(SalesforceModelBase, cls).add_to_class(name, value)  # type: ignore[misc]


if TYPE_CHECKING:
    class SalesforceModel(models.Model, Generic[_T], metaclass=SalesforceModelBase):
        _salesforce_object = ...

        def __init__(self, *args: Any, **kwargs: Any) -> None:  # pylint:disable=super-init-not-called
            tmp = models.manager.Manager()  # type: models.manager.Manager[_T]
            self.objects = tmp

        class Meta:
            ...
else:
    class SalesforceModel(models.Model, metaclass=SalesforceModelBase):
        """
        Abstract model class for Salesforce objects.
        """
        _salesforce_object = 'standard'

        # If both managers are specified in this class, they should be different objects,
        # otherwse the name of the base manager must be passed to the mataclass before
        # the name of default manager, but that order is randomized by Python <= 3.5.x
        base_manager = manager.SalesforceManager()  # type: manager.SalesforceManager[_T]
        objects = manager.SalesforceManager()  # type: manager.SalesforceManager[_T]

        class Meta:
            abstract = True
            base_manager_name = 'base_manager'

        # Name of primary key 'id' can be changed to 'Id' by "settings.SF_PK='id'".
        id = SalesforceAutoField(primary_key=True, name=SF_PK, db_column='Id',
                                 verbose_name='ID', auto_created=True)


class ModelTemplate:
    Meta = SalesforceModel.Meta


def make_dynamic_fields(pattern_module: types.ModuleType, dynamic_field_patterns: Iterable[str],
                        attrs: Dict[str, Any]) -> None:
    """Add some Salesforce fields from a pattern_module models.py

    Parameters:
        pattern_module:  Module where to search additional fields settings.
           It is an imported module created by introspection (inspectdb),
           usually named `models_template.py`. (You will probably not add it
           to version control because the diffs are frequent and huge.)
        dynamic_field_patterns:  List of regular expression for Salesforce
            field names that should be included automatically into the model.
        attrs:  Input/Output dictionary of model attributes. (no need to
            worry, added automatically)

    The patterns are applied sequentionally.
    If the pattern starts with "-" then the matched names are excluded.
    The search stops after the first match.
    A normal field that exists directly in a class is never rewritten
    by a dynamic field..
    All ForeingKey fields should be created explicitely. (For now to
    prevent possible issues and also for better readibility of the
    model. The automatic "dynamic" fields are intended especially for
    "maybe can be useful" fields and will work with ForeignKey in simple
    cases, e.g. without Proxy models etc. Works good for me.)

    This is useful for development: Many fields or all fields can be
    easily accessed by the model without a huge code. Finally
    all wildcard fields except the explicit names can be removed when
    the development is ready or .
    If you create migrations, you probably want to disable "dynamic_field_patterns"
    by setting them empty.

    Example:
        Meta:
            db_table = 'Contact'
            dynamic_patterns = exported.models, ['Last', '.*Date$']
    """
    # pylint:disable=too-many-branches,too-many-locals
    attr_meta = attrs['Meta']
    db_table = getattr(attr_meta, 'db_table', None)
    if not db_table:
        raise RuntimeError('The "db_table" must be set in Meta if "dynamic_field_patterns" is used.')
    is_custom_model = getattr(attr_meta, 'custom', False)

    patterns = []
    for pat in dynamic_field_patterns:
        enabled = True
        if pat.startswith('-'):
            enabled = False
            pat = pat[1:]
        patterns.append((enabled, re.compile(r'^(?:{})$'.format(pat), re.I)))

    used_columns = []
    for name, attr in attrs.items():
        if isinstance(attr, SfField):
            field = attr
            if field.sf_custom is None and is_custom_model:
                field.sf_custom = True
            if not field.name:
                field.name = name
            attname, column = field.get_attname_column()  # pylint:disable=unused-variable
            used_columns.append(column)

    if not pattern_module:
        raise RuntimeError("a pattern_module is required for dynamic fields.")
    for name, obj in vars(pattern_module).items():
        if not name.startswith('_') and isclass(obj) and issubclass(obj, ModelTemplate):
            default_table = obj.__name__
            if getattr(getattr(obj, 'Meta', None), 'db_table', default_table) == db_table:
                cls = obj
                break
    else:
        # not found db_table model, but decide between warning or exception
        if any(not x.startswith('__') for x in dir(pattern_module)):
            raise RuntimeError("No Model for table '%s' found in the module '%s'"
                               % (db_table, pattern_module.__name__))
        warnings.warn("The module '%s' is empty. (It is OK if you are "
                      "rewriting new Models by pipe from inspectdb command.)"
                      % pattern_module.__name__)
        return
    lazy_fields = [(name, obj) for name, obj in vars(cls).items()
                   if isinstance(obj, LazyField) and issubclass(obj.klass, SfField)
                   ]
    for name, obj in sorted(lazy_fields, key=lambda name_obj: name_obj[1].counter):
        for enabled, pattern in patterns:
            if pattern.match(name):
                break
        else:
            enabled = False
        if enabled:
            if issubclass(obj.klass, ForeignKey):
                to = obj.args[0]
                if isclass(to) and issubclass(to, ModelTemplate):
                    obj.args = (to.__name__,) + obj.args[1:]
            field = obj.create()
            attrs[name] = field
    assert pattern_module  # maybe rarely locked while running inspectdb


Model = SalesforceModel



# indep - optional functions and classes that should be independent on the
#         rest of salesforce to not make dependency graphs complicated
import uuid
from inspect import signature
from typing import Any, Callable, Dict, Tuple, Type  # pylint:disable=unused-import # Type

from django.conf import settings
from django.db.models import Field  # pylint:disable=unused-import


class LazyField:
    """A Field that can be later customized until it is binded to the final Model"""
    # It does not need a deserializer for migrations because it is never passed
    # to migration tracking before before activation to a normal field

    counter = 0

    def __init__(self, klass: 'Type[Field[Any, Any]]') -> None:
        """Instantiate the field type"""
        self.klass = klass
        self.kw = {}  # type: Dict[str, Any]
        self.args = ()  # type: Tuple[Any, ...]
        self.called = False
        self.counter = self.counter

    def __call__(self, *args: Any, **kwargs: Any) -> 'LazyField':
        """Instantiate a new field with options"""
        assert not self.called
        # check valid args and check duplicite
        # the method Signature.bind() prefers *args over **kwargs if it is ambiguaous
        bound_args = signature(self.klass.__init__).bind(self, *args, **kwargs)
        obj = type(self)(self.klass)
        obj.args = bound_args.args[1:]
        obj.kw = bound_args.kwargs
        setattr(type(self), 'counter', getattr(type(self), 'counter') + 1)
        return obj

    def update(self, **kwargs: Any) -> 'LazyField':
        """Customize the lazy field"""
        assert not self.called
        self.kw.update(kwargs)
        return self

    def create(self) -> 'Field[Any, Any]':
        """Create a normal field from the lazy field"""
        assert not self.called
        return self.klass(*self.args, **self.kw)


def uuid_pk() -> str:
    return uuid.uuid4().hex


def get_sf_alt_pk() -> Callable[[], str]:
    return getattr(settings, 'SF_ALT_PK', uuid_pk)()  # type: ignore[no-any-return] # noqa



"""
CursorWrapper (like 'django.db.backends.utils' and a half of driver)

It does not inherit from it because e.g.:
- transactions are not supported
- it is not useful to close connection or rollback after error

but some functionality can be moved to the driver and not duplicated
"""
import decimal
import logging
import warnings
from itertools import islice
from typing import Any, Callable, Dict, Iterable, Iterator, List, Tuple, TypeVar, Union, overload

from django.db import models
from django.db.models import expressions as db_expressions
from django.db.models.sql import subqueries, Query, RawQuery

from salesforce.backend import DJANGO_30_PLUS, DJANGO_42_PLUS, DJANGO_50_PLUS
from salesforce.dbapi.driver import (
    DatabaseError, SalesforceWarning, merge_dict,
    register_conversion, arg_to_json)
from salesforce.fields import NOT_UPDATEABLE, NOT_CREATEABLE

if DJANGO_42_PLUS:
    from django.core.exceptions import FullResultSet  # type: ignore[attr-defined] # pylint:disable=ungrouped-imports
else:
    class FullResultSet(Exception):  # type: ignore[no-redef]
        pass


V = TypeVar('V')
if not DJANGO_30_PLUS:
    # a "do nothing" stub for Django < 3.0, where is no decorator @async_unsafe
    F = TypeVar('F', bound=Callable)
    F2 = TypeVar('F2', bound=Callable)

    @overload
    def async_unsafe(message: F) -> F:
        ...

    @overload
    def async_unsafe(message: str) -> Callable[[F2], F2]:
        ...

    def async_unsafe(message: Union[F, str]) -> Union[F, Callable[[F2], F2]]:
        def decorator(func: F2) -> F2:
            return func

        # If the message is actually a function, then be a no-arguments decorator.
        if callable(message):
            func = message
            message = 'You cannot call this from an async context - use a thread or sync_to_async.'
            return decorator(func)
        return decorator
else:

    from django.utils.asyncio import (  # type: ignore[import,no-redef] # noqa pylint:disable=unused-import,ungrouped-imports
        async_unsafe
    )

log = logging.getLogger(__name__)


DJANGO_DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S.%f-00:00'

MIGRATIONS_QUERY_TO_BE_IGNORED = "SELECT django_migrations.app, django_migrations.name FROM django_migrations"


def extract_insert_values(query) -> List[Dict[str, Any]]:  # TODO can be more strict
    """
    Extract values from insert.
    Supports bulk_create
    """
    assert query.model
    ret = []
    for row in query.objs:
        d = dict()
        fields = query.model._meta.fields
        for _, field in enumerate(fields):
            sf_read_only = getattr(field, 'sf_read_only', 0)
            if field.get_internal_type() == 'AutoField':
                continue
            value = getattr(row, field.attname)
            if ((sf_read_only & NOT_CREATEABLE) != 0 or hasattr(value, 'default') or
                    value is None and getattr(field, 'db_default', None) is not None):
                continue  # skip not createable or DEFAULTED_ON_CREATE
            d[field.column] = arg_to_json(value)
        ret.append(d)
    return ret


def extract_update_values(query: subqueries.UpdateQuery) -> Dict[str, Any]:  # TODO can be more strict
    """
    Extract values from update query.
    """
    d = dict()
    assert query.model
    fields = query.model._meta.fields
    for _, field in enumerate(fields):
        sf_read_only = getattr(field, 'sf_read_only', 0)
        is_date_auto = getattr(field, 'auto_now', False) or getattr(field, 'auto_now_add', False)
        if field.get_internal_type() == 'AutoField':
            continue
        if (sf_read_only & NOT_UPDATEABLE) != 0 or is_date_auto:
            continue
        value_or_empty = [value for qfield, model, value in query.values if qfield.name == field.name]
        if value_or_empty:
            [value] = value_or_empty
        else:
            assert len(query.values) < len(fields), \
                "Match name can miss only with an 'update_fields' argument."
            continue
        if hasattr(value, 'default'):
            warnings.warn(
                "The field '{}.{}' has been saved again with DEFAULTED_ON_CREATE value. "
                "It is better to use 'db_default=...' in Django >= 5.0 "
                "or to set a real value to it "
                "or to refresh it from the database after .save() "
                "or to restrict updated fields explicitly by 'update_fields='."
                .format(field.model._meta.object_name, field.name),
                SalesforceWarning
            )
            continue
        d[field.column] = arg_to_json(value)
    return d


class CursorWrapper:
    """
    A wrapper that emulates the behavior of a database cursor.

    This is the class that is actually responsible for making connections
    to the SF REST API
    """

    # pylint:disable=too-many-instance-attributes,too-many-public-methods
    def __init__(self, cursor, db) -> None:
        """
        Connect to the Salesforce API.
        """
        self.cursor = cursor
        self.db = db
        self.query = None
        self.session = db.sf_session  # this creates a TCP connection if doesn't exist
        self.rowcount = None
        self.first_row = None
        self.lastrowid = None  # not moved to driver because INSERT is implemented here

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @property
    def oauth(self):
        return self.session.auth.get_auth()

    def execute(self, q, args=()):
        """
        Send a query to the Salesforce API.
        """
        # pylint:disable=too-many-branches
        self.rowcount = None
        response = None
        if self.query is None:
            self.execute_select(q, args)
        else:
            response = self.execute_django(q, args)
            if isinstance(response, list):
                return

        # the encoding is detected automatically, e.g. from headers
        if response and response.text:
            # parse_float set to decimal.Decimal to avoid precision errors when
            # converting from the json number to a float and then to a Decimal object
            # on a model's DecimalField. This converts from json number directly
            # to a Decimal object
            data = response.json(parse_float=decimal.Decimal)
            # a SELECT query
            if 'totalSize' in data:
                # SELECT
                self.rowcount = data['totalSize']
            # a successful INSERT query, return after getting PK
            elif 'success' in data and 'id' in data:
                self.lastrowid = data['id']
                return
            elif 'compositeResponse' in data:
                # TODO treat error reporting for composite requests
                self.lastrowid = [x['body']['id'] if x['body'] is not None else x['referenceId']
                                  for x in data['compositeResponse']]
                return
            elif data['hasErrors'] is False:
                # it is from Composite Batch request
                # save id from bulk_create even if Django don't use it
                if data['results'] and data['results'][0]['result']:
                    self.lastrowid = [item['result']['id'] for item in data['results']]
                return
            # something we don't recognize
            else:
                raise DatabaseError(data)

            if not q.upper().startswith('SELECT COUNT() FROM'):
                self.first_row = data['records'][0] if data['records'] else None

    def prepare_query(self, query):
        self.query = query

    def execute_django(self, soql: str, args: Tuple[Any, ...] = ()):
        """
        Fixed execute for queries coming from Django query compilers
        """
        response = None
        sqltype = soql.split(None, 1)[0].upper()
        if isinstance(self.query, subqueries.InsertQuery):
            response = self.execute_insert(self.query)
        elif isinstance(self.query, subqueries.UpdateQuery):
            response = self.execute_update(self.query)
        elif isinstance(self.query, subqueries.DeleteQuery):
            response = self.execute_delete(self.query)
        elif isinstance(self.query, RawQuery):
            self.execute_select(soql, args)
        elif sqltype in ('SAVEPOINT', 'ROLLBACK', 'RELEASE'):
            log.info("Ignored SQL command '%s'", sqltype)
            return
        elif isinstance(self.query, Query):
            self.execute_select(soql, args)
        else:
            raise DatabaseError("Unsupported query: type %s: %s" % (type(self.query), self.query))
        return response

    def execute_select(self, soql: str, args) -> None:
        if soql != MIGRATIONS_QUERY_TO_BE_IGNORED:
            # normal query
            query_all = self.query and self.query.sf_params.query_all
            tooling_api = self.query and self.query.model._meta.sf_tooling_api_model
            self.cursor.execute(soql, args, query_all=query_all, tooling_api=tooling_api)
        else:
            # Nothing queried about django_migrations to SFDC and immediately responded that
            # nothing about migration status is recorded in SFDC.
            #
            # That is required by "makemigrations" to accept this query.
            # Empty results are possible.
            # (It could be eventually replaced by: "SELECT app__c, Name FROM django_migrations__c")
            self.cursor._iter = iter([])  # pylint:disable=protected-access
            self.cursor.rowcount = 0
        self.rowcount = self.cursor.rowcount

    def our_fix_default(self, obj_json_data: Dict[str, Any]) -> None:
        if DJANGO_50_PLUS:
            # sql, params = obj_json_data[name].as_sql(self.query.get_compiler('salesforce'), self.db)
            ignore_names = [
                name for name, val in obj_json_data.items()
                if isinstance(val, db_expressions.DatabaseDefault)]  # type: ignore[attr-defined] # ok DJANGO_50_PLUS
            for name in ignore_names:
                del obj_json_data[name]

    def execute_insert(self, query):
        table = query.model._meta.db_table
        post_data = extract_insert_values(query)
        if table == 'django_migrations':
            return
        obj_url = self.db.connection.rest_api_url('sobjects', table, relative=True)
        if len(post_data) == 1:
            # single object
            post_data_0 = post_data[0]
            self.our_fix_default(post_data_0)
            return self.handle_api_exceptions('POST', obj_url, json=post_data_0)
        if self.db.connection.composite_type == 'sobject-collections':
            # SObject Collections
            records = [merge_dict(x, type_=table) for x in post_data]
            for item in records:
                self.our_fix_default(item)
            all_or_none = query.sf_params.all_or_none
            ret = self.db.connection.sobject_collections_request('POST', records, all_or_none=all_or_none)
            self.lastrowid = ret
            self.rowcount = len(ret)
            return
        # composite by REST
        composite_data = [{'method': 'POST', 'url': obj_url, 'referenceId': str(i), 'body': row}
                          for i, row in enumerate(post_data)]
        ret = self.db.connection.composite_request(composite_data)
        return ret

    def get_pks_from_query(self, query):
        """Prepare primary keys for update and delete queries"""
        where = query.where
        sql = None
        if where.connector == 'AND' and not where.negated and len(where.children) == 1:
            # simple cases are optimized, especially because a suboptimal
            # nested query based on the same table is not allowed by SF
            child = where.children[0]
            if (hasattr(child, 'lookup_name') and child.lookup_name in ('exact', 'in')
                    and child.lhs.target.column == 'Id'
                    and not child.bilateral_transforms and child.lhs.target.model is self.query.model):
                pks = child.rhs
                if child.lookup_name == 'exact':
                    assert isinstance(pks, str)
                    return [pks]
                # lookup_name 'in'
                assert not child.bilateral_transforms
                if isinstance(pks, (tuple, list)):
                    return pks
                # 'sf_params' are also in 'pks' only in Django >= 2.0, therefore check query.sf_params
                assert (isinstance(pks, Query) and type(pks).__name__ == 'SalesforceQuery' or
                        query.sf_params.edge_updates), (
                    "Too complicated queryset.update(). Rewrite it by two querysets. "
                    "See docs wiki/error-messages")
                # # alternative solution:
                # return list(salesforce.backend.query.SalesforceQuerySet(pk.model, query=pk, using=pk._db))

                sql, params = pks.get_compiler('salesforce').as_sql()
        if not sql:
            # a subquery is necessary in this case
            try:
                where_sql, params = where.as_sql(query.get_compiler('salesforce'), self.db.connection)
            except FullResultSet:
                where_sql, params = "", []
            sql = "SELECT Id FROM {}".format(query.model._meta.db_table)
            if where_sql:
                sql += " WHERE {}".format(where_sql)
        with self.db.cursor() as cur:
            cur.execute(sql, params)
            assert len(cur.description) == 1 and cur.description[0][0] == 'Id'
            return [x[0] for x in cur]

    def execute_tooling_update(self, query):
        table = query.model._meta.db_table
        post_data = extract_update_values(query)
        pks = self.get_pks_from_query(query)
        assert len(pks) == 1
        pk = pks[0]
        value_map = {qfield.db_column: value for qfield, _, value in query.values}
        if 'Metadata' in value_map and 'FullName' in value_map and 'DurableId' in value_map:
            ret = self.db.connection.handle_api_exceptions(
                'PATCH',
                'tooling/sobjects', table, value_map['DurableId'],
                json={'Metadata': value_map['Metadata'], 'FullName': value_map['FullName']}
            )
        elif pk == '000000000000000AAA':
            pks = value_map['DurableId']
            post_data = dict(**{"attributes": {"type": query.model._meta.db_table}}, **post_data)
            obj_url = self.db.connection.rest_api_url('tooling/sobjects', table, value_map['DurableId'], relative=True)
            ret = self.db.connection.handle_api_exceptions('PATCH', obj_url, json=value_map)
        else:
            obj_url = self.db.connection.rest_api_url('tooling/sobjects', table, pk, relative=True)
            ret = self.db.connection.handle_api_exceptions('PATCH', obj_url, json=post_data)
        assert ret.status_code == 204
        self.rowcount = 1

    def execute_update(self, query):
        if query.model._meta.sf_tooling_api_model:
            return self.execute_tooling_update(query)
        table = query.model._meta.db_table
        post_data = extract_update_values(query)
        pks = self.get_pks_from_query(query)
        log.debug('UPDATE %s(%s)%r', table, pks, post_data)
        if not pks:
            return
        obj_url = self.db.connection.rest_api_url('sobjects', table, '', relative=True)
        if len(pks) == 1:
            # single request
            self.our_fix_default(post_data)
            ret = self.handle_api_exceptions('PATCH', obj_url + pks[0], json=post_data)
            self.rowcount = 1
            return ret
        if self.db.connection.composite_type == 'sobject-collections':
            # SObject Collections
            records = [merge_dict(post_data, id=pk, type_=table) for pk in pks]
            for item in records:
                self.our_fix_default(item)
            all_or_none = query.sf_params.all_or_none
            ret = self.db.connection.sobject_collections_request('PATCH', records, all_or_none=all_or_none)
            self.lastrowid = ret
            self.rowcount = len(ret)
            return
        # composite by REST
        composite_data = [{'method': 'PATCH', 'url': obj_url + pk, 'referenceId': pk, 'body': post_data}
                          for pk in pks]
        ret = self.db.connection.composite_request(composite_data)
        self.rowcount = len([x for x in ret.json()['compositeResponse'] if x['httpStatusCode'] == 204])
        return ret

    def execute_delete(self, query):
        table = query.model._meta.db_table
        pks = self.get_pks_from_query(query)

        log.debug('DELETE %s(%s)', table, pks)
        if not pks:
            self.rowcount = 0
            return
        if len(pks) == 1:
            ret = self.handle_api_exceptions('DELETE', 'sobjects', table, pks[0])
            self.rowcount = 1 if (ret and ret.status_code == 204) else 0
            return ret
        if self.db.connection.composite_type == 'sobject-collections':
            # SObject Collections
            records = pks
            all_or_none = None  # sf_params not supported by DeleteQuery
            ret = self.db.connection.sobject_collections_request('DELETE', records, all_or_none=all_or_none)
            self.lastrowid = ret
            self.rowcount = len(ret)
            return
        # bulk by REST
        url = self.db.connection.rest_api_url('sobjects', table, '', relative=True)
        composite_data = [{'method': 'DELETE', 'url': url + pk, 'referenceId': pk}
                          for pk in pks]
        ret = self.db.connection.composite_request(composite_data)
        self.rowcount = len([x for x in ret.json()['compositeResponse'] if x['httpStatusCode'] == 204])

    def __iter__(self):
        return self.cursor

    def fetchone(self):
        return self.cursor.fetchone()

    def fetchmany(self, size=None):
        return self.cursor.fetchmany(size=size)

    def fetchall(self):
        return self.cursor.fetchall()

    @property
    def description(self):
        return self.cursor.description

    def close(self):
        self.cursor.close()

    def commit(self):
        self.cursor.commit()

    def rollback(self):
        self.cursor.rollback()

    def handle_api_exceptions(self, method, *url_parts, **kwargs):
        return self.cursor.handle_api_exceptions(method, *url_parts, **kwargs)


def chunked(iterable: Iterable[V], n: int) -> Iterator[List[V]]:
    """
    Break an iterable into lists of a given length::

    >>> assert list(chunked([1, 2, 3, 4, 5], 3)) == [[1, 2, 3], [4,5]]
    """
    iterable = iter(iterable)
    while True:
        chunk = list(islice(iterable, n))
        if not chunk:
            return
        yield chunk


def sobj_id(obj):
    assert obj._salesforce_object  # pylint:disable=protected-access
    return obj.pk


# this JSON conversion is important for QuerySet.update(foreign_key_field=some_object)
register_conversion(models.Model, json_conv=sobj_id, subclass=True)



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

from django.contrib import admin
from salesforce.testrunner.example import models
from salesforce.testrunner.example.universal_admin import register_omitted_classes


class AccountAdmin(admin.ModelAdmin):
    # list_display = ('Salutation', 'Name', 'PersonEmail')
    list_display = ('Name', 'Phone')


admin.site.register(models.Account, AccountAdmin)

register_omitted_classes(models)



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""

# from django.test import TestCase



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#
"""
Salesforce Database backend for Django.

No code in this directory is used with standard databases, even if a standard
database is used for running some application tests on objects defined by
SalesforceModel. All code for SF models that can be used with non SF databases
should be located directly in the 'salesforce' directory in files 'models.py',
'fields.py', 'router.py', 'admin.py', 'backend/manager.py'.

Most of code here (in salesforce.backend) is private. (undocumented like other
Django backends and can be changed anytime. Therefore it should not be used by apps)
The only public code here is `salesforce/backend/manager.py`.

Structure:
    salesforce/*.py - Code that is to be used instead of some standard
                      django.db classes.
    salesforce/backend/*.py - Private code equivalent to django.db.backend.some_backend
    salesforce/dbapi/*.py - Database driver independent on Django

Incorrectly located files: (It is better not to change it now.)
    backend/manager.py   => manager.py
    auth.py              => dbapi/auth.py
"""

import logging
import re

import django

DJANGO_30_PLUS = django.VERSION[:2] >= (3, 0)
DJANGO_31_PLUS = django.VERSION[:2] >= (3, 1)
DJANGO_32_PLUS = django.VERSION[:2] >= (3, 2)
DJANGO_40_PLUS = django.VERSION[:2] >= (4, 0)
DJANGO_41_PLUS = django.VERSION[:2] >= (4, 1)
DJANGO_42_PLUS = django.VERSION[:2] >= (4, 2)
DJANGO_50_PLUS = django.VERSION[:2] >= (5, 0)
DJANGO_51_PLUS = django.VERSION[:2] >= (5, 1)
DJANGO_52_PLUS = django.VERSION[:2] >= (5, 2)
max_django = (5, 2)
is_dev_version = bool(django.VERSION[3:] and re.match('(alpha|beta|rc)', django.VERSION[3]))
if django.VERSION[:2] < (2, 2) or django.VERSION[:2] > max_django and not is_dev_version:
    raise ImportError("Django version between 2.2 and 5.2 is required "
                      "for this django-salesforce.")
    # Usually three or more blocking issues can be expected by every
    # new major Django version. Strict check before support is better.

    # New Django development versions are enabled without any restriction, but
    # new stable Django versions must be verified before they are enabled here.


log = logging.getLogger(__name__)



import re
import sys
from django.test import TestCase
import django
from salesforce import backend


class VersionTest(TestCase):
    def test_version(self):
        """Check that the expected Django version is being tested

        "tox" sometimes installed an incorrect Django version,
        e.g. if versions required by tox deps or setup.py are in conflict.
        The unexpected version should be immediately uninstalled and replaced by a requiered version.
        """
        match = re.search(r'/\.tox/dj(\d+|dev)(?:-py(\d+))?/', django.__file__)
        if match:
            django_version_abbr, python_version_abbr = match.groups()
            if django_version_abbr == 'dev':
                self.assertTrue(backend.is_dev_version, "a DEVELOPMENT version of Django should be installed "
                                "in this DEV environment")
                self.assertGreater(django.VERSION[:2], backend.max_django)
            else:
                self.assertEqual('{}{}'.format(*django.VERSION[:2]), django_version_abbr)
            if python_version_abbr is not None:
                self.assertEqual('{}{}'.format(*sys.version_info[:2]), python_version_abbr)



"""Parse SOQL, including subqueries

Subqueries are very restricted by [Force.com SOQL]
(https://resources.docs.salesforce.com/sfdc/pdf/salesforce_soql_sosl.pdf)
(It's lucky for our implementation)
- "only root queries support aggregate expressions"
- Subqueries can not be in "OR" condition, because it is an additional level.

These expressions can be supported here, but currently unimplemented:
    FORMAT(), convertCurrency(), toLabel(),
    Date Functions..., convertTimezone() in Date Functions
    DISTANCE(x, GEOLOCATION(...))
Especially DISTANCE and GEOLOCATION are not trivial because they would
require a combined parser for parentheses and commas.

Unsupported GROUP BY ROLLUP and GROUP BY CUBE (their syntax for reports).
"""
from typing import Any, Callable, Dict, Iterable, List, Optional, overload, Sequence, Type, TypeVar, Tuple, Union
import datetime
import re
import pytz
from salesforce.dbapi.exceptions import ProgrammingError

_TRow = TypeVar('_TRow', Tuple[Any, ...], List[Any], Dict[str, Any])
_T = TypeVar('_T')


def nz(x: Optional[_T]) -> _T:
    """Not Zero type assert"""
    assert x is not None
    return x


# reserved wods can not be used as alias names
RESERVED_WORDS = set((
    'AND, ASC, DESC, EXCLUDES, FIRST, FROM, GROUP, HAVING, '
    'IN, INCLUDES, LAST, LIKE, LIMIT, NOT, NULL, NULLS, '
    'OR, SELECT, WHERE, WITH'
).split(', '))
AGGREGATION_WORDS = set((
    'AVG, COUNT, COUNT_DISTINCT, MIN, MAX, SUM'
).split(', '))
pattern_aggregation = re.compile(r'\b(?:{})(?=\()'.format('|'.join(AGGREGATION_WORDS)), re.I)
pattern_groupby = re.compile(r'\bGROUP BY\b', re.I)


class QQuery:
    """Parse the SOQL query to an object useful to correctly interpret a response.

    public methods:
        parse_rest_response(records, rowcount, row_type=list):

            Parse the response to rows of the tape list or dict
            parameters:
                records: response['records'] or a slice of its iterator
                rowcount: response['totalSize']

            The `cursor` that created the response can be necessary, but currently unused
    """
    # It can be later splitted to more specialized objects, if QQuery become
    # extended and too complicated.
    # type of query: QRootQuery, QFieldSubquery, QWhereSubquery
    # type of field: QField, QAggregation, QFieldSubquery

    # pylint:disable=too-many-instance-attributes

    def __init__(self, soql: Optional[str] = None) -> None:
        self.soql = None                  # type: Optional[str]
        self.fields = []                  # type: List[Union[str, QQuery]]
        # dictionary of chil to parent relationships - lowercase keys
        self.subroots = {}                # type: Dict[str, Any]  # recursive is Any
        self.aliases = []                 # type: List[str]
        self.root_table = None            # type: Optional[str]
        # is_aggregation: only to know if aliases are relevant for output
        self.is_aggregation = False
        self.has_child_rel_field = False
        # extra_soql: everything what is after the root table name (and
        # after optional alias), usually WHERE...
        self.extra_soql = None            # type: Optional[str]
        self.subqueries = None            # type: Optional[List[Tuple[str, Any]]]
        if soql:
            self._from_sql(soql)

    def _from_sql(self, soql: str) -> None:
        """Create Force.com SOQL tree structure from SOQL"""
        # pylint:disable=too-many-branches,too-many-nested-blocks
        assert not self.soql, "Don't use _from_sql method directly"
        self.soql = soql
        soql, self.subqueries = split_subquery(soql)
        match_parse = re.match(r'SELECT (.*) FROM (\w+)\b(.*)$', soql, re.I)
        if not match_parse:
            raise ProgrammingError('Invalid SQL: %s' % self.soql)
        fields_sql, self.root_table, self.extra_soql = match_parse.groups()
        fields = [x.strip() for x in fields_sql.split(',')]
        self.is_aggregation = bool(pattern_groupby.search(self.extra_soql) or
                                   pattern_aggregation.search(fields[0]))
        self.is_plain_count = fields[0].upper() == 'COUNT()'
        consumed_subqueries = 0
        expr_alias_counter = 0
        #
        if not self.is_plain_count:
            out_field = ''  # type: Union[str, QQuery]
            for field in fields:
                if self.is_aggregation:
                    match = re.search(r'\b\w+$', field)
                    if match:
                        alias = match.group()
                        assert alias not in RESERVED_WORDS, "invalid alias name"
                        if match.start() > 0 and field[match.start() - 1] == ' ':
                            out_field = field = field[match.start() - 1]
                    else:
                        alias = 'expr{}'.format(expr_alias_counter)
                        expr_alias_counter += 1
                    assert '&' not in field, "Subquery not expected as field in aggregation query"
                elif '&' in field:
                    assert field == '(&)'  # verify that the subquery was in parentheses
                    subquery = QQuery(self.subqueries[consumed_subqueries][0])
                    consumed_subqueries += 1
                    self.has_child_rel_field = True
                    out_field = subquery
                    # TODO more child relationships to the same table
                    alias = nz(subquery.root_table)
                else:
                    alias = out_field = field
                    if '.' in alias:
                        if alias.split('.', 1)[0].lower() == self.root_table.lower():
                            alias = alias.split('.', 1)[1]
                        if '.' in alias:
                            # prepare paths for possible empty outer joins
                            subroots = self.subroots
                            root_crumbs = alias.lower().split('.')[:-1]
                            for scrumb in root_crumbs:
                                subroots.setdefault(scrumb, {})
                                subroots = subroots[scrumb]
                self.aliases.append(alias)
                self.fields.append(out_field)
        # TODO it is not currently necessary to parse the exta_soql

    def _make_flat(self, row_dict: Dict[str, Any], path: Tuple[str, ...], subroots: Dict[str, Dict[str, Any]]
                   ) -> Dict[str, Any]:
        """Replace the nested dict objects by a flat dict with keys "object.object.name"."""
        # can get a cursor parameter, if introspection should be possible on the fly
        out = {}
        for k, v in row_dict.items():
            klc = k.lower()  # "key lower case"
            if (not (isinstance(v, dict) and 'attributes' in v)
                    or ('done' in v and 'records' in v and 'totalSize' in v)):
                # :
                if klc not in subroots:
                    out[klc] = v
                else:
                    strpath = '.'.join(path + (klc,)) + '.'
                    strip_pos = len(strpath) - len(klc + '.')
                    for alias in self.aliases:
                        if alias.lower().startswith(strpath):
                            out[alias.lower()[strip_pos:]] = None  # empty outer join field names
            else:
                new_subroots = subroots[klc] if k != 'attributes' else {}
                for sub_k, sub_v in self._make_flat(v, path + (klc,), new_subroots).items():
                    out[k.lower() + '.' + sub_k] = sub_v
        return out

    @overload                   # noqa
    def parse_rest_response(self, records: Iterable[Dict[str, Any]], rowcount: int, row_type: Type[_TRow]
                            ) -> Iterable[_TRow]: ...
    @overload                   # noqa
    def parse_rest_response(self, records: Iterable[Dict[str, Any]], rowcount: int,
                            ) -> Iterable[Tuple[Any, ...]]: ...
    def parse_rest_response(self, records: Iterable[Dict[str, Any]], # type: ignore[no-untyped-def] # noqa
                            rowcount: int, row_type=tuple) -> Iterable[Any]:
        """Parse the REST API response to DB API cursor flat response"""
        assert row_type in (dict, list, tuple)
        if self.is_plain_count:
            # result of "SELECT COUNT() FROM ... WHERE ..."
            assert list(records) == []
            if issubclass(row_type, dict):
                yield {'count': rowcount}
            elif issubclass(row_type, list):
                yield [rowcount]  # originally [resp.json()['totalSize']]
            elif issubclass(row_type, tuple):
                yield (rowcount,)
        else:
            for row_deep in records:
                if list(row_deep.keys()) == ['explain']:
                    yield (row_deep['explain'],)
                else:
                    assert self.is_aggregation == (row_deep['attributes']['type'] == 'AggregateResult')
                    row_flat = self._make_flat(row_deep, path=(), subroots=self.subroots)
                    # TODO Will be the expression "or x['done']" really correct also for long subrequests?
                    assert all(not isinstance(x, dict) or x['done'] for x in row_flat)
                    if issubclass(row_type, dict):
                        yield {k: fix_data_type(row_flat[k.lower()]) for k in self.aliases}
                    elif issubclass(row_type, list):
                        yield [fix_data_type(row_flat[k.lower()]) for k in self.aliases]
                    elif issubclass(row_type, tuple):
                        yield tuple(fix_data_type(row_flat[k.lower()]) for k in self.aliases)


SALESFORCE_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f+0000'
SF_DATETIME_PATTERN = re.compile(r'[1-3]\d{3}-[01]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-6]\d.\d{3}\+0000$')


def fix_data_type(data: Any, tzinfo: Optional[datetime.timezone] = None) -> Any:
    # this is a simplified function. The data type should be finally read
    # from some reliable field mapping, not to guess by regexp like here.
    # Only a DateTime field has so specific regexp that the guess is
    # acceptable.
    if isinstance(data, str) and SF_DATETIME_PATTERN.match(data):
        datim = datetime.datetime.strptime(data, SALESFORCE_DATETIME_FORMAT)
        datim = datim.replace(tzinfo=tzinfo or pytz.utc)
        return datim
    return data


def mark_quoted_strings(sql: str) -> Tuple[str, List[str]]:
    """Mark all quoted strings in the SOQL by '@' and get them as params,
    with respect to all escaped backslashes and quotes.
    """
    # pattern of a string parameter (pm), a char escaped by backslash (bs)
    # out_pattern: characters valid in SOQL
    pm_pattern = re.compile(r"'[^\\']*(?:\\[\\'][^\\']*)*'")
    bs_pattern = re.compile(r"\\([\\'])")
    out_pattern = re.compile(r"^(?:[-!()*+,.:<=>\w\s|%s])*$")
    missing_apostrophe = "invalid character in SOQL or a missing apostrophe"
    start = 0
    out = []
    params = []
    for match in pm_pattern.finditer(sql):
        out.append(sql[start:match.start()])
        assert out_pattern.match(sql[start:match.start()]), missing_apostrophe
        params.append(bs_pattern.sub('\\1', sql[match.start() + 1:match.end() - 1]))
        start = match.end()
    out.append(sql[start:])
    assert out_pattern.match(sql[start:]), missing_apostrophe
    return '@'.join(out), params


def subst_quoted_strings(sql: str, params: Sequence[str]) -> str:
    """Reverse operation to mark_quoted_strings - substitutes '@' by params.
    """
    parts = sql.split('@')
    params_dont_match = "number of parameters doesn' match the transformed query"
    assert len(parts) == len(params) + 1, params_dont_match  # would be internal error
    out = []
    for i, param in enumerate(params):
        out.append(parts[i])
        out.append("'%s'" % param.replace('\\', '\\\\').replace("\'", "\\\'"))
    out.append(parts[-1])
    return ''.join(out)


def find_closing_parenthesis(sql: str, startpos: int) -> Optional[Tuple[int, int]]:
    """Find the pair of opening and closing parentheses.

    Starts search at the position startpos.
    Returns tuple of positions (opening, closing) if search succeeds, otherwise None.
    """
    pattern = re.compile(r'[()]')
    level = 0
    opening = []  # type: Union[int, List[None]]
    for match in pattern.finditer(sql, startpos):
        par = match.group()
        if par == '(':
            if level == 0:
                opening = match.start()
            level += 1
        if par == ')':
            assert level > 0, "missing '(' before ')'"
            level -= 1
            if level == 0:
                assert isinstance(opening, int)
                closing = match.end()
                return opening, closing
    return None


def transform_except_subquery(sql: str, func: Callable[[str], str]) -> str:
    """Call a func on every part of SOQL query except nested (SELECT ...)"""
    start = 0
    out = []
    while sql.find('(SELECT', start) > -1:
        pos = sql.find('(SELECT', start)
        out.append(func(sql[start:pos]))
        start, pos = nz(find_closing_parenthesis(sql, pos))
        out.append(sql[start:pos])
        start = pos
    out.append(func(sql[start:len(sql)]))
    return ''.join(out)


def split_subquery(sql: str) -> Tuple[str, List[Tuple[str, Any]]]:
    """Split on subqueries and replace them by '&'."""
    sql, params = mark_quoted_strings(sql)
    sql = simplify_expression(sql)
    _ = params  # NOQA
    start = 0
    out = []
    subqueries = []  # List[Tuple[str, Any]]
    pattern = re.compile(r'\(SELECT\b', re.I)
    match = pattern.search(sql, start)
    while match:
        out.append(sql[start:match.start() + 1] + '&')
        start, pos = nz(find_closing_parenthesis(sql, match.start()))
        start, pos = start + 1, pos - 1
        subqueries.append(split_subquery(sql[start:pos]))
        start = pos
        match = pattern.search(sql, start)
    out.append(sql[start:len(sql)])
    return ''.join(out), subqueries


def simplify_expression(txt: str) -> str:
    """Remove all unecessary whitespace and some very usual space"""
    minimal = re.sub(r'\s', ' ',
                     re.sub(r'\s(?=\W)', '',
                            re.sub(r'(?<=\W)\s', '',
                                   txt.strip())))
    # add space before some "(" and after some ")"
    return re.sub(r'\)(?=\w)', ') ',
                  re.sub(r'(,|\b(?:{}))\('.format('|'.join(RESERVED_WORDS)), '\\1 (', minimal)
                  )
